import collections
import dataclasses
import enum
import functools
import json
import logging
import pathlib
import uuid
from datetime import datetime, timezone
from typing import Any, Callable, Collection, Dict, List, Literal, Optional, TypeVar, Union

import hamilton.node
from hamilton import graph_types
from hamilton.caching import fingerprinting
from hamilton.caching.cache_key import create_cache_key
from hamilton.caching.stores.base import (
    MetadataStore,
    ResultRetrievalError,
    ResultStore,
    search_data_adapter_registry,
)
from hamilton.caching.stores.file import FileResultStore
from hamilton.caching.stores.sqlite import SQLiteMetadataStore
from hamilton.function_modifiers.metadata import cache as cache_decorator
from hamilton.graph import FunctionGraph
from hamilton.lifecycle.base import (
    BaseDoNodeExecute,
    BasePostNodeExecute,
    BasePreGraphExecute,
    BasePreNodeExecute,
)

logger = logging.getLogger("hamilton.caching")

SENTINEL = object()
S = TypeVar("S", object, object)


CACHING_BEHAVIORS = Literal["default", "recompute", "disable", "ignore"]


class CachingBehavior(enum.Enum):
    """Behavior applied by the caching adapter

    DEFAULT:
        Try to retrieve result from cache instead of executing the node. If the node is executed, store the result.
        Compute the result data version and store it too.

    RECOMPUTE:
        Don't try to retrieve result from cache and always execute the node. Otherwise, behaves as default.
        Useful when nodes are stochastic (e.g., model training) or interact with external
        components (e.g., read from database).

    DISABLE:
        Node is executed as if the caching feature wasn't enabled.
        It never tries to retrieve results. Results are never stored nor versioned.
        Behaves like IGNORE, but the node remains a dependency for downstream nodes.
        This means downstream cache lookup will likely fail systematically (i.e., if the cache is empty).

    IGNORE:
        Node is executed as if the caching feature wasn't enable.
        It never tries to retrieve results. Results are never stored nor versioned.
        IGNORE means downstream nodes will ignore this node as a dependency for lookup.
        Ignoring clients and connections can be useful since they shouldn't directly impact the downstream results.
    """

    DEFAULT = 1
    RECOMPUTE = 2
    DISABLE = 3
    IGNORE = 4

    @classmethod
    def from_string(cls, string: str) -> "CachingBehavior":
        """Create a caching behavior from a string of the enum value. This is
        leveraged by the ``hamilton.lifecycle.caching.SmartCacheAdapter`` and
        the ``hamilton.function_modifiers.metadata.cache`` decorator.

        .. code-block::

            CachingBehavior.from_string("recompute")

        """
        try:
            return cls[string.upper()]
        except KeyError as e:
            raise KeyError(f"{string} is an invalid `CachingBehavior` value") from e


class NodeRoleInTaskExecution(enum.Enum):
    """Identify the role of a node in task-based execution, in particular when
    ``Parallelizable/Collect`` are used.

    NOTE This is an internal construct and it will likely change in the future.

    STANDARD: when task-based execution is not used. All nodes and dependencies are STANDARD.
    EXPAND: node with type ``Parallelizable``. It returns an iterator where individual items need to be handled.
        Dependencies can only be OUTSIDE.
    COLLECT: node with type ``Collect``. It returns an iterable where individual items need to be handled.
        Dependencies can be INSIDE, OUTSIDE, or EXPAND
    OUTSIDE: "outside" of ``Parallelizable/Collect`` paths; handled like STANDARD in most cases.
        Dependencies can be OUTSIDE or COLLECT
    INSIDE: "inside" or "between" a ``Parallelizable/Collect`` nodes.
        Dependencies can be INSIDE, OUTSIDE, or EXPAND.
    """

    STANDARD = 1
    EXPAND = 2
    COLLECT = 3
    OUTSIDE = 4
    INSIDE = 5


class CachingEventType(enum.Enum):
    """Event types logged by the caching adapter"""

    GET_DATA_VERSION = "get_data_version"
    SET_DATA_VERSION = "set_data_version"
    GET_CACHE_KEY = "get_cache_key"
    SET_CACHE_KEY = "set_cache_key"
    GET_RESULT = "get_result"
    SET_RESULT = "set_result"
    MISSING_RESULT = "missing_result"
    FAILED_RETRIEVAL = "failed_retrieval"
    EXECUTE_NODE = "execute_node"
    FAILED_EXECUTION = "failed_execution"
    RESOLVE_BEHAVIOR = "resolve_behavior"
    UNHASHABLE_DATA_VERSION = "unhashable_data_version"
    IS_OVERRIDE = "is_override"
    IS_INPUT = "is_input"
    IS_FINAL_VAR = "is_final_var"
    IS_DEFAULT_PARAMETER_VALUE = "is_default_parameter_value"


@dataclasses.dataclass(frozen=True)
class CachingEvent:
    """Event logged by the caching adapter"""

    run_id: str
    actor: Literal["adapter", "metadata_store", "result_store"]
    event_type: CachingEventType
    node_name: str
    task_id: Optional[str] = None
    msg: Optional[str] = None
    value: Optional[Any] = None
    timestamp: float = dataclasses.field(
        default_factory=lambda: datetime.now(timezone.utc).timestamp()
    )

    def __str__(self) -> str:
        """Create a human-readable string format for `print()`"""

        string = self.node_name
        if self.task_id is not None:
            string += f"::{self.task_id}"
        string += f"::{self.actor}"
        string += f"::{self.event_type.value}"
        if self.msg:  # this catches None and empty strings
            string += f"::{self.msg}"

        return string

    def as_dict(self):
        return dict(
            run_id=self.run_id,
            timestamp=self.timestamp,
            node_name=self.node_name,
            task_id=self.task_id,
            actor=self.actor,
            event_type=self.event_type.value,
            msg=self.msg,
            value=str(self.value) if self.value else self.value,
        )


# TODO we could add a "driver-level" kwarg to specify the cache format (e.g., parquet, JSON, etc.)
class HamiltonCacheAdapter(
    BaseDoNodeExecute, BasePreGraphExecute, BasePostNodeExecute, BasePreNodeExecute
):
    """Adapter enabling Hamilton's caching feature through ``Builder.with_cache()``

    .. code-block:: python

        from hamilton import driver
        import my_dataflow

        dr = (
            driver.Builder()
            .with_modules(my_dataflow)
            .with_cache()
            .build()
        )

        # then, you can access the adapter via
        dr.cache

    """

    def __init__(
        self,
        path: Union[str, pathlib.Path] = ".hamilton_cache",
        metadata_store: Optional[MetadataStore] = None,
        result_store: Optional[ResultStore] = None,
        default: Optional[Union[Literal[True], Collection[str]]] = None,
        recompute: Optional[Union[Literal[True], Collection[str]]] = None,
        ignore: Optional[Union[Literal[True], Collection[str]]] = None,
        disable: Optional[Union[Literal[True], Collection[str]]] = None,
        default_behavior: Optional[CACHING_BEHAVIORS] = None,
        default_loader_behavior: Optional[CACHING_BEHAVIORS] = None,
        default_saver_behavior: Optional[CACHING_BEHAVIORS] = None,
        log_to_file: bool = False,
        **kwargs,
    ):
        """Initialize the cache adapter.

        :param path: path where the cache metadata and results will be stored
        :param metadata_store: BaseStore handling metadata for the cache adapter
        :param result_store: BaseStore caching dataflow execution results
        :param default: Set caching behavior to DEFAULT for specified node names. If True, apply to all nodes.
        :param recompute: Set caching behavior to RECOMPUTE for specified node names. If True, apply to all nodes.
        :param ignore: Set caching behavior to IGNORE for specified node names. If True, apply to all nodes.
        :param disable: Set caching behavior to DISABLE for specified node names. If True, apply to all nodes.
        :param default_behavior: Set the default caching behavior.
        :param default_loader_behavior: Set the default caching behavior `DataLoader` nodes.
        :param default_saver_behavior: Set the default caching behavior `DataSaver` nodes.
        :param log_to_file: If True, append cache event logs as they happen in JSONL format.
        """
        self._path = path
        self.metadata_store = (
            metadata_store if metadata_store is not None else SQLiteMetadataStore(path=path)
        )
        self.result_store = (
            result_store if result_store is not None else FileResultStore(path=str(path))
        )
        self.log_to_file = log_to_file

        if sum([default is True, recompute is True, disable is True, ignore is True]) > 1:
            raise ValueError(
                "Can only set one of (`default`, `recompute`, `disable`, `ignore`) to True. Please pass mutually exclusive sets of node names"
            )
        self._default = default
        self._recompute = recompute
        self._disable = disable
        self._ignore = ignore
        self.default_behavior = default_behavior
        self.default_loader_behavior = default_loader_behavior
        self.default_saver_behavior = default_saver_behavior

        # attributes populated at execution time
        self.run_ids: List[str] = []
        self._fn_graphs: Dict[str, FunctionGraph] = {}  # {run_id: graph}
        self._data_savers: Dict[str, Collection[str]] = {}  # {run_id: list[node_name]}
        self._data_loaders: Dict[str, Collection[str]] = {}  # {run_id: list[node_name]}
        self.behaviors: Dict[
            str, Dict[str, CachingBehavior]
        ] = {}  # {run_id: {node_name: behavior}}
        self.data_versions: Dict[
            str, Dict[str, Union[str, Dict[str, str]]]
        ] = {}  # {run_id: {node_name: version}} or {run_id: {node_name: {task_id: version}}}
        self.code_versions: Dict[str, Dict[str, str]] = {}  # {run_id: {node_name: version}}
        self.cache_keys: Dict[
            str, Dict[str, Union[str, Dict[str, str]]]
        ] = {}  # {run_id: {node_name: key}} or {run_id: {node_name: {task_id: key}}}
        self._logs: Dict[str, List[CachingEvent]] = {}  # {run_id: [logs]}

    @property
    def last_run_id(self):
        """Run id of the last started run. Not necessarily the last to complete."""
        return self.run_ids[-1]

    def __getstate__(self) -> dict:
        """Serialization method required for multiprocessing and multithreading
        when using task-based execution with `Parallelizable/Collect`
        """
        state = self.__dict__.copy()
        # store the classes to reinstantiate the same backend in __setstate__
        state["metadata_store_cls"] = self.metadata_store.__class__
        state["metadata_store_init"] = self.metadata_store.__getstate__()
        state["result_store_cls"] = self.result_store.__class__
        state["result_store_init"] = self.result_store.__getstate__()
        del state["metadata_store"]
        del state["result_store"]
        return state

    def __setstate__(self, state: dict) -> None:
        """Serialization method required for multiprocessing and multithreading
        when using task-based execution with `Parallelizable/Collect`.

        Create new instances of metadata and result stores to have one connection
        per thread.
        """
        # instantiate the backend from the class, then delete the attribute before
        # setting it on the adapter instance.
        self.metadata_store = state["metadata_store_cls"](**state["metadata_store_init"])
        self.result_store = state["result_store_cls"](**state["result_store_init"])
        del state["metadata_store_cls"]
        del state["result_store_cls"]
        self.__dict__.update(state)

    def _log_event(
        self,
        run_id: str,
        node_name: str,
        actor: Literal["adapter", "metadata_store", "result_store"],
        event_type: CachingEventType,
        msg: Optional[str] = None,
        value: Optional[Any] = None,
        task_id: Optional[str] = None,
    ) -> None:
        """Add a single event to logs stored in state, keyed by run_id

        If global log level is set to logging.INFO, only log if event type is GET_RESULT or EXECUTE_NODE;
        If it is set to logging.DEBUG, log all events.

        If `SmartCacheAdapter.log_to_file` is set to True, log all events to a file in JSONL format.

        :param node_name: name of the node associated with the event
        :param task_id: optional identifier when using task-based execution. (node_name, task_id) is a primary key
        :param actor: component responsible for the event
        :param event_type: enum specifying what type of event (execute, retrieve, etc.)
        :param msg: additional message to display in the logs (e.g., via terminal)
        :param value: arbitrary value to include (typically a string for data version, code version, cache_key). Must be small and JSON-serializable.
        """
        event = CachingEvent(
            run_id=run_id,
            node_name=node_name,
            task_id=task_id,
            actor=actor,
            event_type=event_type,
            msg=msg,
            value=value,
        )
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(f"{event.__str__()}")
        elif logger.isEnabledFor(logging.INFO):
            if event.event_type in (CachingEventType.GET_RESULT, CachingEventType.EXECUTE_NODE):
                logger.info(f"{event.__str__()}")

        self._logs[run_id].append(event)

        if self.log_to_file:
            log_file_path = pathlib.Path(self.metadata_store._directory, "cache_logs.jsonl")
            json_line = json.dumps(event.as_dict())
            with log_file_path.open("a") as f:
                f.write(json_line + "\n")

    def _log_by_node_name(
        self, run_id: str, level: Literal["debug", "info"] = "info"
    ) -> Dict[str, List[str]]:
        """For a given run, group logs to key them by ``node_name`` or ``(node_name, run_id)`` if applicable."""
        run_logs = collections.defaultdict(list)
        for event in self._logs[run_id]:
            if level == "info":
                if event.event_type not in (
                    CachingEventType.GET_RESULT,
                    CachingEventType.EXECUTE_NODE,
                ):
                    continue

            key = (event.node_name, event.task_id) if event.task_id else event.node_name
            run_logs[key].append(event)
        return dict(run_logs)

    def logs(self, run_id: Optional[str] = None, level: Literal["debug", "info"] = "info") -> dict:
        """Execution logs of the cache adapter.

        :param run_id: If ``None``, return all logged runs. If provided a ``run_id``, group logs by node.
        :param level: If ``"debug"`` log all events. If ``"info"`` only log if result is retrieved or executed.
        :return: a mapping between node/task and a list of logged events

        .. code-block:: python

            from hamilton import driver
            import my_dataflow

            dr = driver.Builder().with_modules(my_dataflow).with_cache().build()
            dr.execute(...)
            dr.execute(...)

            all_logs = dr.cache.logs()
            # all_logs is a dictionary with run_ids as keys and lists of CachingEvent as values.
            # {
            #    run_id_1: [CachingEvent(...), CachingEvent(...)],
            #    run_id_2: [CachingEvent(...), CachingEvent(...)],
            # }


            run_logs = dr.cache.logs(run_id=dr.last_run_id)
            # run_logs are keyed by ``node_name``
            # {node_name: [CachingEvent(...), CachingEvent(...)], ...}
            # or ``(node_name, task_id)`` if task-based execution is used.
            # {(node_name_1, task_id_1): [CachingEvent(...), CachingEvent(...)], ...}

        """
        if run_id:
            return self._log_by_node_name(run_id=run_id, level=level)

        logs = collections.defaultdict(list)
        for run_id, run_logs in self._logs.items():
            for event in run_logs:
                if level == "info" and event.event_type not in (
                    CachingEventType.GET_RESULT,
                    CachingEventType.EXECUTE_NODE,
                ):
                    continue

                logs[run_id].append(event)

        return dict(logs)

    @staticmethod
    def _view_run(
        fn_graph: FunctionGraph,
        logs,
        final_vars: List[str],
        inputs: dict,
        overrides: dict,
        output_file_path: Optional[str] = None,
    ):
        """Create a Hamilton visualization of the execution and the cache hits/misses.

        This leverages the ``custom_style_function`` feature internally.
        """
        from hamilton.driver import Driver  # avoid circular import

        def _visualization_styling_function(*, node, node_class, logs):
            """Custom style function for the visualization."""
            if any(
                event.event_type == CachingEventType.GET_RESULT for event in logs.get(node.name, [])
            ):
                style = (
                    {"penwidth": "3", "color": "#F06449", "fillcolor": "#ffffff"},
                    node_class,
                    "from cache",
                )
            else:
                style = ({}, node_class, None)

            return style

        return Driver._visualize_execution_helper(
            adapter=None,
            bypass_validation=True,
            render_kwargs={},
            output_file_path=output_file_path,
            fn_graph=fn_graph,
            final_vars=final_vars,
            inputs=inputs,
            overrides=overrides,
            custom_style_function=functools.partial(_visualization_styling_function, logs=logs),
        )

    # TODO make this work directly from the metadata_store too
    # visualization from logs is convenient when debugging someone else's issue
    def view_run(self, run_id: Optional[str] = None, output_file_path: Optional[str] = None):
        """View the dataflow execution, including cache hits/misses.

        :param run_id: If ``None``, view the last run. If provided a ``run_id``, view that run.
        :param output_file_path: If provided a path, save the visualization to a file.

        .. code-block:: python

            from hamilton import driver
            import my_dataflow

            dr = driver.Builder().with_modules(my_dataflow).with_cache().build()

            # execute 3 times
            dr.execute(...)
            dr.execute(...)
            dr.execute(...)

            # view the last run
            dr.cache.view_run()
            # this is equivalent to
            dr.cache.view_run(run_id=dr.last_run_id)

            # get a specific run id
            run_id = dr.cache.run_ids[1]
            dr.cache.view_run(run_id=run_id)

        """
        if run_id is None:
            run_id = self.last_run_id

        fn_graph = self._fn_graphs[run_id]
        logs = self.logs(run_id, level="debug")

        final_vars = []
        inputs = {}
        overrides = {}
        for key, events in logs.items():
            if isinstance(key, tuple):
                raise ValueError(
                    "`.view()` is currently not supported for task-based execution. "
                    "Please inspect the logs directly via `.logs(run_id=...)` for debugging."
                )

            node_name = key
            if any(e.event_type == CachingEventType.IS_FINAL_VAR for e in events):
                final_vars.append(node_name)

            if any(e.event_type == CachingEventType.IS_INPUT for e in events):
                inputs[node_name] = None  # the value doesn't matter, only the key of the dict
                continue

            elif any(e.event_type == CachingEventType.IS_OVERRIDE for e in events):
                overrides[node_name] = None  # the value doesn't matter, only the key of the dict
                continue

        return self._view_run(
            fn_graph=fn_graph,
            logs=logs,
            final_vars=final_vars,
            inputs=inputs,
            overrides=overrides,
            output_file_path=output_file_path,
        )

    def _get_node_role(
        self, run_id: str, node_name: str, task_id: Optional[str]
    ) -> NodeRoleInTaskExecution:
        """Determine based on the node name and task_id if a node is part of parallel execution."""
        if task_id is None:
            role = NodeRoleInTaskExecution.STANDARD
        else:
            node_type: hamilton.node.NodeType = self._fn_graphs[run_id].nodes[node_name].node_role
            if node_type == hamilton.node.NodeType.EXPAND:
                role = NodeRoleInTaskExecution.EXPAND
            elif node_type == hamilton.node.NodeType.COLLECT:
                role = NodeRoleInTaskExecution.COLLECT
            elif node_name == task_id:
                role = NodeRoleInTaskExecution.OUTSIDE
            else:
                role = NodeRoleInTaskExecution.INSIDE

        return role

    def get_cache_key(
        self, run_id: str, node_name: str, task_id: Optional[str] = None
    ) -> Union[str, S]:
        """Get the ``cache_key`` stored in-memory for a specific ``run_id``, ``node_name``, and ``task_id``.

        This method is public-facing and can be used directly to inspect the cache.

        :param run_id: Id of the Hamilton execution run.
        :param node_name: Name of the node associated with the cache key. ``node_name`` is a unique identifier
            if task-based execution is not used.
        :param task_id: Id of the task when task-based execution is used. Then, the tuple ``(node_name, task_id)``
            is a unique identifier.
        :return: The cache key if it exists, otherwise return a sentinel value.

        .. code-block:: python

            from hamilton import driver
            import my_dataflow

            dr = driver.Builder().with_modules(my_dataflow).with_cache().build()
            dr.execute(...)

            dr.cache.get_cache_key(run_id=dr.last_run_id, node_name="my_node", task_id=None)

        """
        node_role = self._get_node_role(run_id=run_id, node_name=node_name, task_id=task_id)

        if node_role == NodeRoleInTaskExecution.INSIDE:
            cache_key = self.cache_keys[run_id].get(node_name, {}).get(task_id, SENTINEL)  # type: ignore ; `task_id` can't be None
        else:
            cache_key = self.cache_keys[run_id].get(node_name, SENTINEL)

        cache_key = cache_key if cache_key is not SENTINEL else None
        self._log_event(
            run_id=run_id,
            node_name=node_name,
            task_id=task_id,
            actor="adapter",
            event_type=CachingEventType.GET_CACHE_KEY,
            msg="hit" if cache_key is not SENTINEL else "miss",
            value=cache_key,
        )
        return cache_key

    def _set_cache_key(
        self, run_id: str, node_name: str, cache_key: str, task_id: Optional[str] = None
    ) -> None:
        """Set the ``cache_key`` stored in-memory for a specific ``run_id``, ``node_name``, and ``task_id``.

        When calling this method, ``cache_key`` must not be ``None``.
        """
        assert cache_key is not None
        node_role = self._get_node_role(run_id=run_id, node_name=node_name, task_id=task_id)
        if node_role in (
            NodeRoleInTaskExecution.STANDARD,
            NodeRoleInTaskExecution.OUTSIDE,
            NodeRoleInTaskExecution.EXPAND,
            NodeRoleInTaskExecution.COLLECT,
        ):
            self.cache_keys[run_id][node_name] = cache_key
        elif node_role == NodeRoleInTaskExecution.INSIDE:
            if self.cache_keys[run_id].get(node_name, SENTINEL) is SENTINEL:
                self.cache_keys[run_id][node_name] = {}
            self.cache_keys[run_id][node_name][task_id] = cache_key  # type: ignore ; we just initialized the nested dict
        else:
            raise ValueError(
                f"Received `{node_role}`. Unhandled `NodeRoleInTaskExecution`, please report this bug."
            )

        self._log_event(
            run_id=run_id,
            node_name=node_name,
            task_id=task_id,
            actor="adapter",
            event_type=CachingEventType.SET_CACHE_KEY,
            value=cache_key,
        )

    def _get_memory_data_version(
        self, run_id: str, node_name: str, task_id: Optional[str] = None
    ) -> Union[str, S]:
        """Get the ``data_version`` stored in-memory for a specific ``run_id``, ``node_name``, and ``task_id``.

        The behavior depends on the ``CacheBehavior`` (e.g., RECOMPUTE, IGNORE, DISABLE, DEFAULT) and
        the ``NodeRoleInTaskExecution`` of the node (e.g., STANDARD, OUTSIDE, INSIDE, EXPAND, COLLECT).

        :param run_id: Id of the Hamilton execution run.
        :param node_name: Name of the node associated with the cache key. ``node_name`` is a unique identifier
            if task-based execution is not used.
        :param task_id: Id of the task when task-based execution is used. Then, the tuple ``(node_name, task_id)``
            is a unique identifier.
        """
        node_role = self._get_node_role(run_id=run_id, node_name=node_name, task_id=task_id)
        if node_role in (
            NodeRoleInTaskExecution.STANDARD,
            NodeRoleInTaskExecution.OUTSIDE,
            NodeRoleInTaskExecution.COLLECT,
        ):
            data_version = self.data_versions[run_id].get(node_name, SENTINEL)
        elif node_role == NodeRoleInTaskExecution.EXPAND:
            data_version = SENTINEL
        elif node_role == NodeRoleInTaskExecution.INSIDE:
            tasks_data_versions = self.data_versions[run_id].get(node_name, SENTINEL)
            if isinstance(tasks_data_versions, dict):
                data_version = tasks_data_versions.get(task_id, SENTINEL)
            else:
                data_version = SENTINEL
        else:
            raise ValueError(
                f"Received `{node_role}`. Unhandled `NodeRoleInTaskExecution`, please report this bug."
            )

        self._log_event(
            run_id=run_id,
            node_name=node_name,
            task_id=task_id,
            actor="adapter",
            event_type=CachingEventType.GET_DATA_VERSION,
            msg="hit" if data_version is not SENTINEL else "miss",
        )
        return data_version

    def _get_stored_data_version(
        self, run_id: str, node_name: str, cache_key: str, task_id: Optional[str] = None
    ) -> Union[str, S]:
        """Get the ``data_version`` stored in the metadata store associated with the ``cache_key``.

        The ``run_id``, ``node_name``, and ``task_id`` are included only for logging purposes.
        """
        data_version = self.metadata_store.get(cache_key=cache_key)
        data_version = SENTINEL if data_version is None else data_version
        self._log_event(
            run_id=run_id,
            node_name=node_name,
            task_id=task_id,
            actor="metadata_store",
            event_type=CachingEventType.GET_DATA_VERSION,
            msg="hit" if data_version is not SENTINEL else "miss",
        )

        return data_version

    def get_data_version(
        self,
        run_id: str,
        node_name: str,
        cache_key: Optional[str] = None,
        task_id: Optional[str] = None,
    ) -> Union[str, S]:
        """Get the ``data_version``  for a specific ``run_id``, ``node_name``, and ``task_id``.

        This method is public-facing and can be used directly to inspect the cache. This will check data versions
        stored both in-memory and in the metadata store.

        :param run_id: Id of the Hamilton execution run.
        :param node_name: Name of the node associated with the data version. ``node_name`` is a unique identifier
            if task-based execution is not used.
        :param task_id: Id of the task when task-based execution is used. Then, the tuple ``(node_name, task_id)``
            is a unique identifier.
        :return: The data version if it exists, otherwise return a sentinel value.

        ..code-block:: python

            from hamilton import driver
            import my_dataflow

            dr = driver.Builder().with_modules(my_dataflow).with_cache().build()
            dr.execute(...)

            dr.cache.get_data_version(run_id=dr.last_run_id, node_name="my_node", task_id=None)

        """

        data_version = self._get_memory_data_version(
            run_id=run_id, node_name=node_name, task_id=task_id
        )

        if data_version is SENTINEL and cache_key is not None:
            data_version = self._get_stored_data_version(
                run_id=run_id, node_name=node_name, task_id=task_id, cache_key=cache_key
            )

        return data_version

    def _set_memory_metadata(
        self, run_id: str, node_name: str, data_version: str, task_id: Optional[str] = None
    ) -> None:
        """Set in-memory data_version whether a task_id is specified or not"""
        assert data_version is not None
        node_role = self._get_node_role(run_id=run_id, node_name=node_name, task_id=task_id)
        if node_role in (
            NodeRoleInTaskExecution.STANDARD,
            NodeRoleInTaskExecution.OUTSIDE,
            NodeRoleInTaskExecution.COLLECT,
        ):
            self.data_versions[run_id][node_name] = data_version
        elif node_role == NodeRoleInTaskExecution.EXPAND:
            self.data_versions[run_id][node_name] = {}
        elif node_role == NodeRoleInTaskExecution.INSIDE:
            if self.data_versions[run_id].get(node_name, SENTINEL) is SENTINEL:
                self.data_versions[run_id][node_name] = {}
            self.data_versions[run_id][node_name][task_id] = data_version  # type: ignore ; we just initialized the nested dict
        else:
            raise ValueError(
                f"Received `{node_role}`. Unhandled `NodeRoleInTaskExecution`, please report this bug."
            )

        self._log_event(
            run_id=run_id,
            node_name=node_name,
            task_id=task_id,
            actor="adapter",
            event_type=CachingEventType.SET_DATA_VERSION,
            value=data_version,
        )

    def _set_stored_metadata(
        self,
        run_id: str,
        node_name: str,
        cache_key: str,
        data_version: str,
        task_id: Optional[str] = None,
    ) -> None:
        """Set data_version in the metadata store associated with the cache_key"""
        self.metadata_store.set(
            run_id=run_id,
            node_name=node_name,
            code_version=self.code_versions[run_id][node_name],
            data_version=data_version,
            cache_key=cache_key,
        )
        self._log_event(
            run_id=run_id,
            node_name=node_name,
            task_id=task_id,
            actor="metadata_store",
            event_type=CachingEventType.SET_DATA_VERSION,
            value=data_version,
        )

    def _version_data(
        self, node_name: str, run_id: str, result: Any, task_id: Optional[str] = None
    ) -> str:
        """Create a unique data version for the result"""
        data_version = fingerprinting.hash_value(result)
        if data_version == fingerprinting.UNHASHABLE:
            self._log_event(
                run_id=run_id,
                node_name=node_name,
                task_id=task_id,
                actor="adapter",
                event_type=CachingEventType.UNHASHABLE_DATA_VERSION,
                msg=f"unhashable type {type(result)}; set CachingBehavior.IGNORE to silence warning",
                value=data_version,
            )
            logger.warning(
                f"Node `{node_name}` has unhashable result of type `{type(result)}`. "
                "Set `CachingBehavior.IGNORE` or register a versioning function to silence warning. "
                "Learn more: https://hamilton.dagworks.io/en/latest/concepts/caching/#caching-behavior\n"
            )
            # if the data version is unhashable, we need to set a random suffix to the cache_key
            # to prevent the cache from thinking this value is constant, causing a cache hit.
            data_version = "<unhashable>" + f"_{uuid.uuid4()}"

        return data_version

    def version_data(self, result: Any, run_id: str = None) -> str:
        """Create a unique data version for the result

        This is a user-facing method.
        """
        # stuff the internal function call to not log event
        return self._version_data(result=result, run_id=run_id, node_name=None)

    def version_code(self, node_name: str, run_id: Optional[str] = None) -> str:
        """Create a unique code version for the source code defining the node"""
        run_id = self.last_run_id if run_id is None else run_id
        node = self._fn_graphs[run_id].nodes[node_name]
        return graph_types.HamiltonNode.from_node(node).version  # type: ignore

    def _execute_node(
        self,
        run_id: str,
        node_name: str,
        node_callable: Callable,
        node_kwargs: Dict[str, Any],
        task_id: Optional[str] = None,
    ) -> Any:
        """Simple wrapper that logs the regular execution of a node."""
        logger.debug(node_name)
        result = node_callable(**node_kwargs)
        self._log_event(
            run_id=run_id,
            node_name=node_name,
            task_id=task_id,
            actor="adapter",
            event_type=CachingEventType.EXECUTE_NODE,
        )
        return result

    @staticmethod
    def _resolve_node_behavior(
        node: hamilton.node.Node,
        default: Optional[Collection[str]] = None,
        disable: Optional[Collection[str]] = None,
        recompute: Optional[Collection[str]] = None,
        ignore: Optional[Collection[str]] = None,
        default_behavior: CACHING_BEHAVIORS = "default",
        default_loader_behavior: CACHING_BEHAVIORS = "default",
        default_saver_behavior: CACHING_BEHAVIORS = "default",
    ) -> CachingBehavior:
        """Determine the cache behavior of a node.
        Behavior specified via the ``Builder`` has precedence over the ``@cache`` decorator.
        Otherwise, set the ``DEFAULT`` behavior.
        If the node is `Parallelizable` enforce the ``RECOMPUTE`` behavior to ensure
        yielded items are versioned individually.
        """
        if node.node_role == hamilton.node.NodeType.EXPAND:
            return CachingBehavior.RECOMPUTE

        behavior_from_tag = node.tags.get(cache_decorator.BEHAVIOR_KEY, SENTINEL)
        if behavior_from_tag is not SENTINEL:
            behavior_from_tag = CachingBehavior.from_string(behavior_from_tag)

        behavior_from_driver = SENTINEL
        for behavior, node_set in (
            (CachingBehavior.DEFAULT, default),
            (CachingBehavior.DISABLE, disable),
            (CachingBehavior.RECOMPUTE, recompute),
            (CachingBehavior.IGNORE, ignore),
        ):
            # guard against default None value
            if node_set is None:
                continue

            if node.name in node_set:
                if behavior_from_driver is not SENTINEL:
                    raise ValueError(
                        f"Multiple caching behaviors specified by Driver for node: {node.name}"
                    )
                behavior_from_driver = behavior

        if behavior_from_driver is not SENTINEL:
            return behavior_from_driver
        elif behavior_from_tag is not SENTINEL:
            return behavior_from_tag
        elif node.tags.get("hamilton.data_loader"):
            return CachingBehavior.from_string(default_loader_behavior)
        elif node.tags.get("hamilton.data_saver"):
            return CachingBehavior.from_string(default_saver_behavior)
        else:
            return CachingBehavior.from_string(default_behavior)

    def resolve_behaviors(self, run_id: str) -> Dict[str, CachingBehavior]:
        """Resolve the caching behavior for each node based on the ``@cache`` decorator
        and the ``Builder.with_cache()`` parameters for a specific ``run_id``.

        This is a user-facing method.

        Behavior specified via ``Builder.with_cache()`` have precedence. If no parameters are specified,
        the ``CachingBehavior.DEFAULT`` is used. If a node is ``Parallelizable`` (i.e., ``@expand``),
        the ``CachingBehavior`` is set to ``CachingBehavior.RECOMPUTE`` to ensure the yielded items
        are versioned individually. Internally, this uses the ``FunctionGraph`` stored for each ``run_id`` and logs
        the resolved caching behavior for each node.

        :param run_id: Id of the Hamilton execution run.
        :return: A dictionary of ``{node name: caching behavior}``.
        """
        graph = self._fn_graphs[run_id]

        _default = self._default
        _disable = self._disable
        _recompute = self._recompute
        _ignore = self._ignore

        if _default is True:
            _default = [n.name for n in graph.get_nodes()]
        elif _disable is True:
            _disable = [n.name for n in graph.get_nodes()]
        elif _recompute is True:
            _recompute = [n.name for n in graph.get_nodes()]
        elif _ignore is True:
            _ignore = [n.name for n in graph.get_nodes()]

        default_behavior = "default"
        if self.default_behavior is not None:
            default_behavior = self.default_behavior

        default_loader_behavior = default_behavior
        if self.default_loader_behavior is not None:
            default_loader_behavior = self.default_loader_behavior

        default_saver_behavior = default_behavior
        if self.default_saver_behavior is not None:
            default_saver_behavior = self.default_saver_behavior

        behaviors = {}
        for node in graph.get_nodes():
            behavior = HamiltonCacheAdapter._resolve_node_behavior(
                node=node,
                default=_default,
                disable=_disable,
                recompute=_recompute,
                ignore=_ignore,
                default_behavior=default_behavior,
                default_loader_behavior=default_loader_behavior,
                default_saver_behavior=default_saver_behavior,
            )
            behaviors[node.name] = behavior

            self._log_event(
                run_id=run_id,
                node_name=node.name,
                task_id=None,
                actor="adapter",
                event_type=CachingEventType.RESOLVE_BEHAVIOR,
                value=behavior,
            )

        # need to handle materializers via a second pass to copy the behavior
        # of their "main node"
        for node in graph.get_nodes():
            if node.tags.get("hamilton.data_loader") is True:
                main_node = node.tags["hamilton.data_loader.node"]
                if main_node == node.name:
                    continue

                # solution for `@dataloader` and `from_`
                if behaviors.get(main_node, None) is not None:
                    behaviors[node.name] = behaviors[main_node]
                # this hacky section is required to support @load_from and provide
                # a unified pattern to specify behavior from the module or the driver
                else:
                    behaviors[node.name] = HamiltonCacheAdapter._resolve_node_behavior(
                        # we create a fake node, only its name matters
                        node=hamilton.node.Node(
                            name=main_node,
                            typ=str,
                            callabl=lambda: None,
                            tags=node.tags.copy(),
                        ),
                        default=_default,
                        disable=_disable,
                        recompute=_recompute,
                        ignore=_ignore,
                        default_behavior=default_loader_behavior,
                    )

                self._data_loaders[run_id].append(main_node)

            if node.tags.get("hamilton.data_saver", None) is not None:
                self._data_savers[run_id].append(node.name)

        return behaviors

    def resolve_code_versions(
        self,
        run_id: str,
        final_vars: Optional[List[str]] = None,
        inputs: Optional[Dict[str, Any]] = None,
        overrides: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, str]:
        """Resolve the code version for each node for a specific ``run_id``.

        This is a user-facing method.

        If ``final_vars`` is None, all nodes will be versioned. If ``final_vars`` is provided,
        the ``inputs`` and ``overrides`` are used to determine the execution path and only
        version the code for these nodes.

        :param run_id: Id of the Hamilton execution run.
        :param final_vars: Nodes requested for execution.
        :param inputs: Input node values.
        :param overrides: Override node values.
        :return: A dictionary of ``{node name: code version}``.
        """
        graph = self._fn_graphs[run_id]

        final_vars = [] if final_vars is None else final_vars
        inputs = {} if inputs is None else inputs
        overrides = {} if overrides is None else overrides

        node_selection = graph.get_nodes()
        if len(final_vars) > 0:
            all_nodes, user_defined_nodes = graph.get_upstream_nodes(final_vars, inputs, overrides)
            node_selection = set(all_nodes) - set(user_defined_nodes)

        return {
            node.name: self.version_code(run_id=run_id, node_name=node.name)
            for node in node_selection
        }

    def _process_input(self, run_id: str, node_name: str, value: Any) -> None:
        """Process input nodes to version data and code.

        To enable caching, input values must be versioned. Since inputs have no associated code,
        set a constant "code version" ``f"input__{node_name}"`` that uniquely identifies this input.
        """
        data_version = self._version_data(node_name=node_name, run_id=run_id, result=value)
        self.code_versions[run_id][node_name] = f"input__{node_name}"
        self.data_versions[run_id][node_name] = data_version
        self._log_event(
            run_id=run_id,
            node_name=node_name,
            task_id=None,
            actor="adapter",
            event_type=CachingEventType.IS_INPUT,
            value=data_version,
        )

    def _process_override(self, run_id: str, node_name: str, value: Any) -> None:
        """Process override nodes to version data and code.

        To enable caching, override values must be versioned. As opposed to executed nodes,
        code and data versions for overrides are not stored because their value is user provided
        and isn't necessarily tied to the code.
        """
        data_version = self._version_data(node_name=node_name, run_id=run_id, result=value)
        self.data_versions[run_id][node_name] = data_version
        self._log_event(
            run_id=run_id,
            node_name=node_name,
            task_id=None,
            actor="adapter",
            event_type=CachingEventType.IS_OVERRIDE,
            value=data_version,
        )

    @staticmethod
    def _resolve_default_parameter_values(
        node_: hamilton.node.Node, node_kwargs: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        If a node uses the function's default parameter values, they won't be part of the
        node_kwargs. To ensure a consistent `cache_key` we want to retrieve default parameter
        values if they're used
        """
        resolved_kwargs = node_kwargs.copy()
        for param_name, param_value in node_.default_parameter_values.items():
            # if the `param_name` not in `node_kwargs`, it means the node uses the default
            # parameter value
            if param_name not in node_kwargs.keys():
                resolved_kwargs.update(**{param_name: param_value})

        return resolved_kwargs

    def pre_graph_execute(
        self,
        *,
        run_id: str,
        graph: FunctionGraph,
        final_vars: List[str],
        inputs: Dict[str, Any],
        overrides: Dict[str, Any],
    ):
        """Set up the state of the adapter for a new execution.

        Most attributes need to be keyed by run_id to prevent potential conflicts because
        the same adapter instance is shared between across all ``Driver.execute()`` calls.
        """
        self.run_ids.append(run_id)
        self.metadata_store.initialize(run_id)
        self._logs[run_id] = []

        self._fn_graphs[run_id] = graph
        self.data_versions[run_id] = {}
        self.cache_keys[run_id] = {}
        self.code_versions[run_id] = self.resolve_code_versions(
            run_id=run_id, final_vars=final_vars, inputs=inputs, overrides=overrides
        )
        # the empty `._data_loaders` and `._data_savers` need to be instantiated before calling
        # `self.resolve_behaviors` because it appends to them
        self._data_loaders[run_id] = []
        self._data_savers[run_id] = []
        self.behaviors[run_id] = self.resolve_behaviors(run_id=run_id)

        # final vars are logged to be retrieved by the ``.view_run()`` method
        for final_var in final_vars:
            self._log_event(
                run_id=run_id,
                node_name=final_var,
                task_id=None,
                actor="adapter",
                event_type=CachingEventType.IS_FINAL_VAR,
            )

        if inputs:
            for node_name, value in inputs.items():
                self._process_input(run_id, node_name, value)

        if overrides:
            for node_name, value in overrides.items():
                self._process_override(run_id, node_name, value)

    def pre_node_execute(
        self,
        *,
        run_id: str,
        node_: hamilton.node.Node,
        kwargs: Dict[str, Any],
        task_id: Optional[str] = None,
        **future_kwargs,
    ):
        """Before node execution or retrieval, create the cache_key and set it in memory.
        The cache_key is created based on the node's code version and its dependencies' data versions.

        Collecting ``data_version`` for upstream dependencies requires handling special cases when
        task-based execution is used:
        - If the current node is ``COLLECT`` , the dependency annotated with ``Collect[]`` needs to
        be versioned item by item instead of versioning the full container. This is because the
        collect order is inconsistent.
        - If the current node is ``INSIDE`` and the dependency is ``EXPAND``, this means the
        ``kwargs`` dictionary contains a single item. We need to version this individual item because
        it will not be available from "inside" the branch for some executors (multiprocessing, multithreading)
        because they lose access to the data_versions of ``OUTSIDE`` nodes stored in ``self.data_versions``.

        """
        node_name = node_.name
        node_kwargs = HamiltonCacheAdapter._resolve_default_parameter_values(node_, kwargs)

        if self.behaviors[run_id][node_name] == CachingBehavior.IGNORE:
            return

        # won't need the cache_key for either result retrieval or storage
        if self.behaviors[run_id][node_name] == CachingBehavior.DISABLE:
            return

        node_role = self._get_node_role(run_id=run_id, node_name=node_name, task_id=task_id)
        collected_name = (
            node_.collect_dependency if node_role == NodeRoleInTaskExecution.COLLECT else SENTINEL
        )

        dependencies_data_versions = {}
        for dep_name, dep_value in node_kwargs.items():
            # resolve caching behaviors
            if self.behaviors[run_id][dep_name] == CachingBehavior.IGNORE:
                # setting the data_version to "<ignore>" in the cache_key means that
                # the value of the dependency appears constant to this node
                dependencies_data_versions[dep_name] = "<ignore>"
                continue
            elif self.behaviors[run_id][dep_name] == CachingBehavior.DISABLE:
                # setting the data_version to "<disable>" with a random suffix in the
                # cache_key means the current node will be a cache miss and forced to recompute
                dependencies_data_versions[dep_name] = "<disable>" + f"_{uuid.uuid4()}"
                continue

            # resolve NodeRoleInTaskExecution
            if task_id is None:
                dep_role = NodeRoleInTaskExecution.STANDARD
            else:
                # want to check if dependency is an EXPAND node. We must not pass the current `task_id`
                dep_role = self._get_node_role(
                    run_id=run_id, node_name=dep_name, task_id="<placeholder>"
                )

            # if dep_role == NodeRoleInTaskExecution.STANDARD:

            if dep_name == collected_name:
                # the collected value should be hashed based on the items, not the container
                items_data_versions = [self.version_data(item, run_id=run_id) for item in dep_value]
                dep_data_version = fingerprinting.hash_value(sorted(items_data_versions))

            elif dep_role == NodeRoleInTaskExecution.EXPAND:
                # if the dependency is `EXPAND`, the kwarg received is a single item yielded by the iterator
                # rather than the full iterable. We must version it directly, similar to a top-level input
                dep_data_version = self.version_data(dep_value, run_id=run_id)

            else:
                tasks_data_versions = self._get_memory_data_version(
                    run_id=run_id, node_name=dep_name, task_id=None
                )
                if tasks_data_versions is SENTINEL:
                    dep_data_version = self.version_data(dep_value, run_id=run_id)
                elif isinstance(tasks_data_versions, dict):
                    dep_data_version = tasks_data_versions.get(task_id)
                else:
                    dep_data_version = tasks_data_versions

            if dep_data_version == fingerprinting.UNHASHABLE:
                # if the data version is unhashable, we need to set a random suffix to the cache_key
                # to prevent the cache from thinking this value is constant, causing a cache hit.
                dep_data_version = "<unhashable>" + f"_{uuid.uuid4()}"

            dependencies_data_versions[dep_name] = dep_data_version

        # create cache_key before execution; will be reused during and after execution
        cache_key = create_cache_key(
            node_name=node_name,
            code_version=self.code_versions[run_id][node_name],
            dependencies_data_versions=dependencies_data_versions,
        )
        self._set_cache_key(
            run_id=run_id, node_name=node_name, task_id=task_id, cache_key=cache_key
        )

    def do_node_execute(
        self,
        *,
        run_id: str,
        node_: hamilton.node.Node,
        kwargs: Dict[str, Any],
        task_id: Optional[str] = None,
        **future_kwargs,
    ):
        """Try to retrieve stored result from previous executions or execute the node.

        Use the previously created cache_key to retrieve the data_version from memory or the metadata_store.
        If data_version is retrieved try to retrieve the result. If it fails, execute the node.
        Else, execute the node.
        """
        node_name = node_.name
        node_callable = node_.callable
        node_kwargs = HamiltonCacheAdapter._resolve_default_parameter_values(node_, kwargs)

        if self.behaviors[run_id][node_name] in (
            CachingBehavior.DISABLE,
            CachingBehavior.IGNORE,
            CachingBehavior.RECOMPUTE,
        ):
            result = self._execute_node(
                run_id=run_id,
                node_name=node_name,
                node_callable=node_callable,
                node_kwargs=node_kwargs,
                task_id=task_id,
            )
            if self.behaviors[run_id][node_name] in (
                CachingBehavior.RECOMPUTE,
                CachingBehavior.IGNORE,
            ):
                cache_key = self.get_cache_key(run_id=run_id, node_name=node_name, task_id=task_id)

                # nodes collected in `._data_loaders` return tuples of (result, metadata)
                # where metadata often includes a timestamp. To ensure we provide a consistent
                # `data_version` / hash, we only hash the result part of the materializer return
                # value and discard the metadata.
                if node_name in self._data_loaders[run_id] and isinstance(result, tuple):
                    result = result[0]

                data_version = self._version_data(node_name=node_name, run_id=run_id, result=result)

                # nodes collected in `._data_savers` return a dictionary of metadata
                # this metadata often includes a timestamp, leading to an unstable hash.
                # we do not version nor store the metadata. This node is executed for its
                # external effect of saving a file
                if node_name in self._data_savers[run_id]:
                    data_version = f"{node_name}__metadata"

                self._set_memory_metadata(
                    run_id=run_id, node_name=node_name, task_id=task_id, data_version=data_version
                )
                self._set_stored_metadata(
                    run_id=run_id,
                    node_name=node_name,
                    task_id=task_id,
                    cache_key=cache_key,
                    data_version=data_version,
                )

            return result

        # cache_key is set in `pre_node_execute`
        cache_key = self.get_cache_key(run_id=run_id, node_name=node_name, task_id=task_id)
        # retrieve data version from memory or metadata_store
        data_version = self.get_data_version(
            run_id=run_id, node_name=node_name, task_id=task_id, cache_key=cache_key
        )

        need_to_compute_node = False
        if data_version is SENTINEL:
            # must execute: data_version not found in memory or in metadata_store
            need_to_compute_node = True
        elif data_version == fingerprinting.UNHASHABLE:
            # must execute: the retrieved data_version is UNHASHABLE, therefore it isn't stored.
            need_to_compute_node = True
        elif self.result_store.exists(data_version) is False:
            # must execute: data_version retrieved, but result store can't find result
            need_to_compute_node = True
            self._log_event(
                run_id=run_id,
                node_name=node_name,
                task_id=task_id,
                actor="result_store",
                event_type=CachingEventType.MISSING_RESULT,
                value=data_version,
            )
        else:
            # try to retrieve: data_version retrieve, result store found result
            try:
                # successful retrieval: retrieve the result; potentially load using the DataLoader if e.g.,``@cache(format="json")``
                result = self.result_store.get(data_version=data_version)
                self._log_event(
                    run_id=run_id,
                    node_name=node_name,
                    task_id=task_id,
                    actor="result_store",
                    event_type=CachingEventType.GET_RESULT,
                    msg="hit",
                    value=data_version,
                )
                # set the data_version previously retrieved (could be from memory or store)
                self._set_memory_metadata(
                    run_id=run_id, node_name=node_name, task_id=task_id, data_version=data_version
                )
            except ResultRetrievalError:
                # failed retrieval: despite finding the result, probably failed loading data using DataLoader if e.g.,``@cache(format="json")``
                self.metadata_store.delete(cache_key=cache_key)
                self.result_store.delete(data_version)
                need_to_compute_node = True

        if need_to_compute_node is True:
            result = self._execute_node(
                run_id=run_id,
                node_name=node_name,
                node_callable=node_callable,
                node_kwargs=node_kwargs,
                task_id=task_id,
            )

            # nodes collected in `._data_loaders` return tuples of (result, metadata)
            # where metadata often includes a timestamp. To ensure we provide a consistent
            # `data_version` / hash, we only hash the result part of the materializer return
            # value and discard the metadata.
            if node_name in self._data_loaders[run_id] and isinstance(result, tuple):
                result = result[0]

            data_version = self._version_data(node_name=node_name, run_id=run_id, result=result)

            # nodes collected in `._data_savers` return a dictionary of metadata
            # this metadata often includes a timestamp, leading to an unstable hash.
            # we do not version nor store the metadata. This node is executed for its
            # external effect of saving a file
            if node_name in self._data_savers[run_id]:
                data_version = f"{node_name}__metadata"

            self._set_memory_metadata(
                run_id=run_id, node_name=node_name, task_id=task_id, data_version=data_version
            )
            self._set_stored_metadata(
                run_id=run_id,
                node_name=node_name,
                task_id=task_id,
                cache_key=cache_key,
                data_version=data_version,
            )

        return result

    def post_node_execute(
        self,
        *,
        run_id: str,
        node_: hamilton.node.Node,
        result: Optional[str],
        success: bool = True,
        error: Optional[Exception] = None,
        task_id: Optional[str] = None,
        **future_kwargs,
    ):
        """Get the cache_key and data_version stored in memory (respectively from
        pre_node_execute and do_node_execute) and store the result in result_store
        if it doesn't exist.
        """
        node_name = node_.name

        if success is False:
            self._log_event(
                run_id=run_id,
                node_name=node_name,
                task_id=task_id,
                actor="adapter",
                event_type=CachingEventType.FAILED_EXECUTION,
                msg=f"{error}",
            )
            return

        if self.behaviors[run_id][node_name] in (
            CachingBehavior.DEFAULT,
            CachingBehavior.RECOMPUTE,
            CachingBehavior.IGNORE,
        ):
            cache_key = self.get_cache_key(run_id=run_id, node_name=node_name, task_id=task_id)
            data_version = self.get_data_version(
                run_id=run_id, node_name=node_name, task_id=task_id, cache_key=cache_key
            )
            assert data_version is not SENTINEL

            # TODO clean up this logic
            # check if a materialized file exist before writing results
            # when using `@cache(format="json")`
            cache_format = (
                self._fn_graphs[run_id]
                .nodes[node_name]
                .tags.get(cache_decorator.FORMAT_KEY, SENTINEL)
            )
            if cache_format is not SENTINEL:
                saver_cls, loader_cls = search_data_adapter_registry(
                    name=cache_format, type_=type(result)
                )  # type: ignore
                materialized_path = self.result_store._materialized_path(data_version, saver_cls)
                materialized_path_missing = not materialized_path.exists()
            else:
                saver_cls, loader_cls = None, None
                materialized_path_missing = False

            result_missing = not self.result_store.exists(data_version)
            if result_missing or materialized_path_missing:
                self.result_store.set(
                    data_version=data_version,
                    result=result,
                    saver_cls=saver_cls,
                    loader_cls=loader_cls,
                )
                self._log_event(
                    run_id=run_id,
                    node_name=node_name,
                    task_id=task_id,
                    actor="result_store",
                    event_type=CachingEventType.SET_RESULT,
                    value=data_version,
                )



import abc
import collections
import enum
import logging
from typing import Any, Dict, List, Optional

from hamilton.execution.grouping import NodeGroupPurpose, TaskImplementation, TaskSpec

logger = logging.getLogger(__name__)


class TaskState(enum.Enum):
    QUEUED = "queued"
    UNINITIALIZED = "uninitialized"
    INITIALIZED = "initialized"
    RUNNING = "running"
    SUCCESSFUL = "successful"
    FAILED = "failed"

    @staticmethod
    def is_terminal(task_state: "TaskState") -> bool:
        return task_state in [TaskState.SUCCESSFUL, TaskState.FAILED]


# TODO -- determine a better set of states for the graph
GraphState = TaskState


class ResultCache(abc.ABC):
    """Cache of intermediate results. Will likely want to add pruning to this..."""

    @abc.abstractmethod
    def write(
        self,
        results: Dict[str, Any],
        group_id: Optional[str] = None,
        spawning_task_id: Optional[str] = None,
    ):
        """Writes results to the cache. This is called after a task is run.

        :param results:  Results to write
        :param task_type: Task type to write -- this is inclucded in case a task returns generators
        """
        pass

    @abc.abstractmethod
    def read(
        self,
        keys: List[str],
        group_id: Optional[str] = None,
        spawning_task_id: Optional[str] = None,
        optional: bool = False,
    ) -> Dict[str, Any]:
        """Reads results in bulk from the cache.

        :param spawning_task_id: Task ID of the task that spawned the task
        whose results you wish to query
        :param group_id: Group ID of the task whose results you wish to query
        :param keys: Keys to access
        :param optional: If true, will allow keys not to be present
        :return: Dictionary of key -> result
        """
        pass


class DictBasedResultCache(ResultCache):
    """Cache of intermediate results. Will likely want to add pruning to this..."""

    def __init__(self, cache: Dict[str, Any]):
        self.cache = cache

    def _format_key(
        self, group_id: Optional[str], spawning_task_id: Optional[str], key: str
    ) -> str:
        return ":".join([item for item in [spawning_task_id, group_id, key] if item is not None])

    def write(
        self,
        results: Dict[str, Any],
        group_id: Optional[str] = None,
        spawning_task_id: Optional[str] = None,
    ):
        results_with_key_assigned = {
            self._format_key(group_id, spawning_task_id, key): value
            for key, value in results.items()
        }
        self.cache.update(results_with_key_assigned)

    def read(
        self,
        keys: List[str],
        group_id: Optional[str] = None,
        spawning_task_id: Optional[str] = None,
        optional: bool = False,
    ) -> Dict[str, Any]:
        """Reads results in bulk from the cache. If its optional, we don't mind if its not there.

        :param keys: Keys to read
        :param group_id: ID of the group (namespace) under which this key will be stored
        :param spawning_task_id: ID of the spawning task (other part of the namespace) under which this
        key will be stored
        :param optional: If true, we don't mind if the key is not there
        :return:  Dictionary of key -> result
        """

        out = {}
        for key in keys:
            formatted_key = self._format_key(group_id, spawning_task_id, key)
            if formatted_key not in self.cache:
                if optional:
                    continue
                else:
                    raise KeyError(f"Key {formatted_key} not found in cache")  # noqa E713
            out[key] = self.cache[formatted_key]
        return out


class ExecutionState:
    """Stores the basic execution state of a DAG. This is responsible for two things:
    1. Be the source of truth of the execution state
    2. Tell us what task to run next
    3. Prep the task for execution (give it the results it needs)
    """

    def __init__(self, tasks: List[TaskSpec], result_cache: ResultCache, run_id: str):
        """Initializes an ExecutionState to all uninitialized. TBD if we want to add in an initialization
        step that can, say, read from a db.


        :param task_graph: Graph of tasks to run.
        :param results_cache:  Cache of results
        """
        # Pool of available tasks. Note these get added dynamically as
        # tasks are run, in case we want to split out
        self.result_cache = result_cache
        self.run_id = run_id
        self.task_states = collections.defaultdict(lambda: TaskState.UNINITIALIZED)
        self.task_pool = {}
        self._initialize_task_pool(tasks)
        self.base_task_pool = {}
        self._initialize_base_task_pool(tasks)
        self.task_queue = collections.deque()
        self._initialize_task_queue()
        self.is_initialized = True
        self.base_reverse_dependencies = self.compute_reverse_dependencies(tasks)

    @staticmethod
    def compute_reverse_dependencies(tasks: List[TaskSpec]) -> Dict[str, List[TaskSpec]]:
        """Computes dependencies in reverse order, E.G. what tasks depend on this task.

        :param tasks:
        :return:
        """
        reverse_dependencies = {task.base_id: [] for task in tasks}
        for task in tasks:
            for dependency in task.base_dependencies:
                reverse_dependencies[dependency].append(task)
        return reverse_dependencies

    def _initialize_task_pool(self, tasks: List[TaskSpec]):
        """Initializes the task pool to all nodes that have no dependencies.

        :param tasks:
        :return:
        """
        for task in tasks:
            # Note that this will break with nested tasks...
            # We want to add every task that is initially spawned and "static"
            # We want to add dynamic tasks later
            if len(task.base_dependencies) == 0:
                self.realize_task(task, None, None, None)

    def _initialize_base_task_pool(self, tasks: List[TaskSpec]):
        """Initializes the base task pool to all nodes.

        :param tasks:
        :return:
        """
        for task in tasks:
            self.base_task_pool[task.base_id] = task

    def _initialize_task_queue(self):
        """Initializes the task queue to all nodes that have no dependencies.

        :param tasks:
        :return:
        """
        for task in self.task_pool.values():
            if len(task.base_dependencies) == 0:
                self.task_queue.append(task)
                self.task_states[task.task_id] = TaskState.QUEUED

    def realize_task(
        self,
        task_spec: TaskSpec,
        spawning_task: Optional[str],
        group_id: Optional[str],
        dependencies: Dict[str, List[str]] = None,
        bind: Dict[str, Any] = None,
    ):
        """Creates a task and enqueues it to the internal queue. This takes a task in "Plan" state
        (E.G. a TaskSpec), and trasnforms it into execution-ready state, freezing the dependencies.

        You can think of a TaskSpec/plan as a function (or class), and an TaskImplementation as
        an instance of that function/class.

        :param base_task:
        :return:
        """

        realized_dependencies = (
            dependencies
            if dependencies is not None
            else {task_dep: [task_dep] for task_dep in task_spec.base_dependencies}
        )
        task_implementation = TaskImplementation(
            spawning_task_base_id=spawning_task,
            base_id=task_spec.base_id,
            nodes=task_spec.nodes,
            purpose=task_spec.purpose,
            outputs_to_compute=task_spec.outputs_to_compute,
            overrides=task_spec.overrides,
            adapter=task_spec.adapter,
            base_dependencies=task_spec.base_dependencies,
            spawning_task_id=spawning_task,
            group_id=group_id,
            # TODO -- actually get these correct...
            # This just assigns the realized dependencies to be the base dependencies
            realized_dependencies=realized_dependencies,
            run_id=self.run_id,
        )
        if bind is not None:
            task_implementation = task_implementation.bind(bind)
        self.task_pool[task_implementation.task_id] = task_implementation
        self.task_states[task_implementation.task_id] = TaskState.INITIALIZED
        return task_implementation

    def realize_parameterized_group(
        self,
        spawning_task_id: str,
        parameterizations: Dict[str, Any],
        input_to_parameterize: str,
    ) -> List[TaskImplementation]:
        """Parameterizes an unordered expand group. These are tasks that are all part of the same
        group. For every result in the list, the input of the next task gets the result.

        :param task:
        :param results:
        :return:
        """
        # first we get all future nodes in the group
        out = []
        tasks_to_repeat = [
            task
            for task in self.base_task_pool.values()
            if task.spawning_task_base_id == spawning_task_id
            and not task.purpose.is_expander()
            and not task.purpose.is_gatherer()
        ]
        task_names_in_group = [task.base_id for task in tasks_to_repeat]
        # Then we parameterize them, making the following changes:
        # 1. We replace base dependencies with the actual task ids
        # 2. We replace the inputs of the first task with the results
        # 3. We replace the inputs of the collect task with the outputs of the other tasks, grouped
        # expander_node = completed_task.get_expander_node()
        # # This should be a list of all the parameterizations
        # results_to_bind = list(result_cache.read([expander_node.name]))
        # For every result in the list, we create a new set of tasks
        name_maps = {}
        # Go through every task we need to repeat
        for group_name, result in parameterizations.items():
            new_tasks = []
            # We create a new set of tasks, and add them to the task pool
            name_map = {}
            for task in tasks_to_repeat:
                # We create a new task, with the result bound to the input
                new_task = self.realize_task(
                    task, spawning_task_id, group_name, bind={input_to_parameterize: result}
                )
                # new_task = new_task.bind({expander_node.name: result})
                # should bind...
                name_map[task.base_id] = new_task.task_id
                new_tasks.append(new_task)
            for new_task in new_tasks:
                # We replace the dependencies with the new task ids
                new_task.realized_dependencies = {
                    dependency: name_map[dependency] if dependency in name_map else [dependency]
                    for dependency in new_task.base_dependencies
                }
                # We add the new tasks to the task pool
                out.append(new_task)
            name_maps[group_name] = name_map
        collector_tasks = [
            task
            for task in self.base_task_pool.values()
            if task.purpose.is_gatherer() and task.spawning_task_base_id == spawning_task_id
        ]

        # Go through every task we need to collect those
        for collector_task in collector_tasks:
            # collector_node = completed_task.get_collector_node()
            # collect_input = collector_node.collect_dependency
            # We create it with no spawning tasks as we don't want to name it differently
            # We should really give it a better unique ID? This is a little hacky
            new_task = self.realize_task(collector_task, None, None)
            new_dependencies = {}
            for dependency in new_task.base_dependencies:
                new_dependencies[dependency] = []
                if dependency in task_names_in_group:
                    for _group_name, name_map in name_maps.items():
                        new_dependencies[dependency].append(name_map[dependency])
                else:
                    new_dependencies[dependency].append(dependency)
            new_task.realized_dependencies = new_dependencies
            out.append(new_task)
        return out

    def write_task_results(self, writer: TaskImplementation, results: Dict[str, Any]):
        results_to_write = results
        if writer.purpose.is_expander():
            # In this case we need to write each result individually
            result_name_for_expansion = writer.get_expander_node().name
            result_for_expansion = list(results_to_write.pop(result_name_for_expansion))
            self.result_cache.write({result_name_for_expansion: result_for_expansion}, None, None)
        # write the rest with the appropriate namespace
        self.result_cache.write(results_to_write, writer.group_id, writer.spawning_task_id)

    def update_task_state(
        self, task_id: str, new_state: TaskState, results: Optional[Dict[str, Any]]
    ):
        """Updates the state of a task based on an external push.
        This also determines which tasks to create/enqueue next. If the node finished succesfully,
        this has two steps:

        Node Creation (turning a TaskSpec into a TaskImplementation)
        1. If its a "expand" node, we first create the children tasks, then the collect that depends
        on them
        2. Else, we create subsequent tasks (unless they are a "collect" node, in which case they
            should be created by the expand node that spawns them)

        Node Enqueuing (putting the task into the queue)
        We then enqueue all tasks that (a) depend on them and (b) only have dependencies that are
        now complete.
        :param task_id: ID of task to update state of
        :param new_state: New state to update to.
        """

        self.task_states[task_id] = new_state
        if not TaskState.is_terminal(new_state):
            return
        logger.debug(
            f"Received update for task {task_id} with state {new_state.value}, "
            f"current queue is {self._format_task_queue()}"
        )
        for item in self.task_queue:
            assert self.task_states[item.task_id] == TaskState.QUEUED, (
                f"Task {item.task_id} is in the queue but not queued, "
                f"state is {self.task_states[item.task_id].value}"
            )
        # Creating task implementations
        task_to_update = self.task_pool[task_id]
        if new_state == TaskState.SUCCESSFUL:
            self.write_task_results(task_to_update, results)
            completed_task: TaskImplementation = self.task_pool[task_id]  # First look up the task
            if completed_task.purpose == NodeGroupPurpose.EXPAND_UNORDERED:
                # In this case we need to extract all the tasks we need to spawn
                # This involves the following:
                #    1. For each node, spawn the task, and bind the result to that node
                #    2. Spawn the collect task to depend on all of them that it depends on)
                input_to_parameterize = completed_task.get_expander_node().name
                # We do realization here, which is not ideal, as it outputs a generator The task
                # executor should be able to handle this I think... Likely we want the executor
                # to be able to write its own result cache? In case it has an internal buffer?
                parameterization_results = self.result_cache.read([input_to_parameterize])[
                    input_to_parameterize
                ]
                parameterization_values = {
                    str(i): item for i, item in enumerate(parameterization_results)
                }
                logger.debug(
                    f"Completed an expand step, parameterizing: {input_to_parameterize} "
                    f"over values : {parameterization_values}"
                )
                self.realize_parameterized_group(
                    completed_task.task_id, parameterization_values, input_to_parameterize
                )
            else:
                for candidate_task in self.base_reverse_dependencies[completed_task.base_id]:
                    # This means its not spawned by another task, or a node spawning group itself
                    if (
                        candidate_task.spawning_task_base_id is None
                        or candidate_task.purpose.is_expander()
                    ):
                        # Then we need to spawn it
                        # In this case it will have no namespace/group (just standard)
                        new_task_id = candidate_task.base_id
                        if new_task_id not in self.task_pool:
                            # Then we need to create it
                            self.realize_task(candidate_task, None, None)
            # TODO -- make this more efficient
            # We need reverse dependencies, but there are ways to do this cleanly
            # For now we'll just loop through

            tasks_to_enqueue = []
            # not efficient, TODO -- use a reverse dependency map
            for _key, task in self.task_pool.items():
                if self.task_states[task.task_id] == TaskState.INITIALIZED:
                    should_launch = True
                    for _base_dep_name, realized_dep_list in task.realized_dependencies.items():
                        for dep in realized_dep_list:
                            if self.task_states[dep] != TaskState.SUCCESSFUL:
                                should_launch = False
                    if should_launch:
                        tasks_to_enqueue.append(task)
            task_names = [task.task_id for task in tasks_to_enqueue]
            if len(task_names) > 0:
                logger.info(
                    f"Enqueuing {len(task_names)} task{'s' if len(task_names) > 1 else ''}: "
                    f"{', '.join(task_names[:2] + (['...'] if len(task_names) > 2 else []))}"
                )
            for task_to_enqueue in tasks_to_enqueue:
                self.task_queue.append(task_to_enqueue)
                self.task_states[task_to_enqueue.task_id] = TaskState.QUEUED

    def _format_task_queue(self, verbose: bool = False) -> str:
        if verbose:
            return str([item.task_id for item in self.task_queue])
        state_counts = collections.Counter(
            [self.task_states[item.task_id].value for item in self.task_queue]
        )
        return str(dict(state_counts))

    def get_graph_state(self) -> GraphState:
        """Gives the state of the graph, which can be derived from the state of the tasks.

        :return: State of the graph
        """
        if not self.is_initialized:
            return GraphState.UNINITIALIZED
        elif len(self.task_states) == 0 or all(
            [state == TaskState.INITIALIZED for state in self.task_states.values()]
        ):
            return GraphState.INITIALIZED
        elif all([state == TaskState.SUCCESSFUL for state in self.task_states.values()]):
            return GraphState.SUCCESSFUL
        elif any([state == TaskState.FAILED for state in self.task_states.values()]):
            # if we want to execute until completion, then we should return RUNNING here
            # Unless they're all terminal, in which case we should return FAILED
            return GraphState.FAILED
        else:
            return GraphState.RUNNING

    def bind_task(self, task: TaskImplementation):
        required_input_vars, optional_input_vars = task.get_input_vars()
        dynamic_inputs = task.overrides.copy()
        if task.purpose.is_gatherer():
            # We need to intelligently bind the arguments from its dependencies
            collector_node = task.get_collector_node()
            collector_arg = collector_node.collect_dependency
            dynamic_inputs = {collector_arg: []}
            for dependency_base, dependencies in task.realized_dependencies.copy().items():
                # We know that if one has the collector node, they all will
                # As that's how we run it
                if len(dependencies) == 0:
                    # This is a parameterized task, so we'll just end up passing an empty array, and not computing anything from it
                    continue
                produces_collector_arg = self.task_pool[dependencies[0]].produces(collector_arg)
                if produces_collector_arg:
                    for dependency in dependencies:
                        dependent_task = self.task_pool[dependency]
                        dynamic_inputs[collector_arg].append(
                            self.result_cache.read(
                                [collector_arg],
                                dependent_task.group_id,
                                dependent_task.spawning_task_id,
                            )[collector_arg]
                        )
                    del task.realized_dependencies[dependency_base]
        # Then we go through the rest -- these just fill in the other args
        input_vars_to_read = [
            item
            for item in required_input_vars
            if item not in dynamic_inputs and item not in task.overrides
        ]
        dynamic_inputs = {**dynamic_inputs, **self.result_cache.read(input_vars_to_read)}
        dynamic_inputs = {
            **dynamic_inputs,
            **self.result_cache.read(optional_input_vars, optional=True),
        }
        return task.bind(dynamic_inputs)

    def release_next_task(self) -> Optional[TaskImplementation]:
        """Gives the next task to run, and inserts inputs it needs to run.
        Note that this can return None, which means there is nothing to run. That indicates that,
        either:

            1. The graph is done computing
            2. All tasks are currently blocked

        The caller is responsible for no longer calling this when get_graph_state() is in a
        terminal position.

        TODO -- consider finalizing (not binding) a task to save compute/external resources.

        :return: None if there is no task
        """
        if len(self.task_queue) == 0:
            return None
        return self.bind_task(self.task_queue.popleft())

    def reject_task(self, task_to_reject: TaskImplementation):
        """If the task just released is not accepted, we're just going to place
        it back on the front of the queue. We'll likely want to redo these, but
        for now this will be OK.

        :param task_to_reject: Task we got from release_next_task that we're rejecting.
        """
        self.task_queue.appendleft(task_to_reject)



import dbm


class JsonCache:
    def __init__(self, cache_path: str):
        self.cache_path = f"{cache_path}/json_cache"

    def keys(self) -> list:
        with dbm.open(self.cache_path, "r") as db:
            return list(db.keys())

    def write(self, data: str, id_: str) -> None:
        with dbm.open(self.cache_path, "c") as db:
            db[id_] = data

    def read(self, id_: str) -> str:
        with dbm.open(self.cache_path, "r") as db:
            return db[id_].decode()

    def delete(self, id_: str) -> None:
        with dbm.open(self.cache_path, "r") as db:
            del db[id_]



import logging

from . import (
    adapters,
    base,
    configuration,
    delayed,
    dependencies,
    expanders,
    macros,
    metadata,
    recursive,
    validation,
)

logger = logging.getLogger(__name__)

"""
Annotations for modifying the way functions get added to the DAG.
All user-facing annotation classes are lowercase as they're meant to be used
as annotations. They are classes to hold state and subclass common functionality.
"""

# These all represent the public API for function_modifiers
# All new user-facing decorators/helper functions should be here

# Backwards-compatibility to be safe
InvalidDecoratorException = base.InvalidDecoratorException

# The config decorator
config = configuration.config
hamilton_exclude = configuration.hamilton_exclude()

# Dependency Specification
# Helper functions to specify dependency sources for parameterization
value = dependencies.value
source = dependencies.source
group = dependencies.group

# These aren't strictly part of the API but we should have them here for safety
LiteralDependency = dependencies.LiteralDependency
UpstreamDependency = dependencies.UpstreamDependency

# Parameterization decorators (both the old and new ones)
# The three "blessed" @parameterize decorators
parameterize = expanders.parameterize
parameterize_sources = expanders.parameterize_sources
parameterize_values = expanders.parameterize_values
parameterize_extract_columns = expanders.parameterize_extract_columns
ParameterizedExtract = expanders.ParameterizedExtract
inject = expanders.inject

# The older ones that will be deprecated
parametrized = expanders.parametrized
parameterized_inputs = expanders.parameterized_inputs
parametrized_input = expanders.parametrized_input

# Extract decorators
extract_columns = expanders.extract_columns
extract_fields = expanders.extract_fields

# does decorator
does = macros.does
pipe = macros.pipe
pipe_input = macros.pipe_input
pipe_output = macros.pipe_output
mutate = macros.mutate
step = macros.step
apply_to = macros.apply_to

# resolve transform/model decorator
dynamic_transform = macros.dynamic_transform
model = macros.model

# Metadata-specifying decorators
tag = metadata.tag
tag_outputs = metadata.tag_outputs
schema = metadata.schema
cache = metadata.cache

# data quality + associated tags
check_output = validation.check_output
check_output_custom = validation.check_output_custom
IS_DATA_VALIDATOR_TAG = validation.IS_DATA_VALIDATOR_TAG
DATA_VALIDATOR_ORIGINAL_OUTPUT_TAG = validation.DATA_VALIDATOR_ORIGINAL_OUTPUT_TAG

# recursive/subdag operators

subdag = recursive.subdag
parameterized_subdag = recursive.parameterized_subdag

# resolve/meta stuff -- power user features

resolve = delayed.resolve
ResolveAt = delayed.ResolveAt
resolve_from_config = delayed.resolve_from_config

# materialization stuff
load_from = adapters.load_from
save_to = adapters.save_to
dataloader = adapters.dataloader
datasaver = adapters.datasaver



import inspect
import sys
from typing import Any, Literal, TypeVar

import numpy as np
import numpy.typing as npt
import pytest

from hamilton import node
from hamilton.node import DependencyType, Node, matches_query

from tests.resources import nodes_with_future_annotation


def test_node_from_fn_happy():
    def fn() -> int:
        return 1

    node = Node.from_fn(fn)
    assert node.name == "fn"
    assert node.callable() == 1
    assert node.input_types == {}
    assert node.type == int


def test_node_from_fn_sad_no_type_hint():
    def fn():
        return 1

    with pytest.raises(ValueError):
        Node.from_fn(fn)


def test_node_copy_with_retains_originating_functions():
    def fn() -> int:
        return 1

    node = Node.from_fn(fn)
    node_copy = node.copy_with(name="rename_fn", originating_functions=(fn,))
    assert node_copy.originating_functions == (fn,)
    assert node_copy.name == "rename_fn"
    node_copy_copy = node_copy.copy_with(name="rename_fn_again")
    assert node_copy_copy.originating_functions == (fn,)
    assert node_copy_copy.name == "rename_fn_again"


@pytest.mark.skipif(sys.version_info < (3, 9), reason="requires python 3.9 or higher")
def test_node_handles_annotated():
    from typing import Annotated

    DType = TypeVar("DType", bound=np.generic)
    ArrayN = Annotated[npt.NDArray[DType], Literal["N"]]

    def annotated_func(first: ArrayN[np.float64], other: float = 2.0) -> ArrayN[np.float64]:
        return first * other

    node = Node.from_fn(annotated_func)
    assert node.name == "annotated_func"
    expected = {
        "first": (
            Annotated[np.ndarray[Any, np.dtype[np.float64]], Literal["N"]],
            DependencyType.REQUIRED,
        ),
        "other": (float, DependencyType.OPTIONAL),
    }
    assert node.input_types == expected
    assert node.type == Annotated[np.ndarray[Any, np.dtype[np.float64]], Literal["N"]]


@pytest.mark.parametrize(
    "tags, query, expected",
    [
        ({}, {"module": "tests.resources.tagging"}, False),
        ({"module": "tests.resources.tagging"}, {}, True),
        ({"module": "tests.resources.tagging"}, {"module": "tests.resources.tagging"}, True),
        ({"module": "tests.resources.tagging"}, {"module": None}, True),
        ({"module": "tests.resources.tagging"}, {"module": None, "tag2": "value"}, False),
        (
            {"module": "tests.resources.tagging"},
            {"module": "tests.resources.tagging", "tag2": "value"},
            False,
        ),
        ({"tag1": ["tag_value1"]}, {"tag1": "tag_value1"}, True),
        ({"tag1": ["tag_value1"]}, {"tag1": ["tag_value1"]}, True),
        ({"tag1": ["tag_value1"]}, {"tag1": ["tag_value1", "tag_value2"]}, True),
        ({"tag1": "tag_value1"}, {"tag1": ["tag_value1", "tag_value2"]}, True),
        ({"tag1": "tag_value1"}, {"tag1": ["tag_value3", "tag_value4"]}, False),
        ({"tag1": ["tag_value1"]}, {"tag1": "tag_value2"}, False),
        ({"tag1": "tag_value1"}, {"tag1": "tag_value2"}, False),
        ({"tag1": ["tag_value1"]}, {"tag1": ["tag_value2"]}, False),
    ],
    ids=[
        "no tags fail",
        "no query pass",
        "exact match pass",
        "match tag key pass",
        "missing extra tag fail",
        "missing extra tag2 fail",
        "list single match pass",
        "list list match pass",
        "list list match one of pass",
        "single list match one of pass",
        "single list fail",
        "list single fail",
        "single single fail",
        "list list fail",
    ],
)
def test_tags_match_query(tags: dict, query: dict, expected: bool):
    assert matches_query(tags, query) == expected


def test_from_parameter_default_override_equals():
    class BrokenEquals:
        def __eq__(self, other):
            raise ValueError("I'm broken")

    def foo(b: BrokenEquals = BrokenEquals()):  # noqa
        pass

    param = DependencyType.from_parameter(inspect.signature(foo).parameters["b"])
    assert param == DependencyType.OPTIONAL


# Tests parsing for future annotations
# TODO -- we should generalize this but doing this for specific points is OK for now
def test_node_from_future_annotation_parallelizable():
    parallelized = nodes_with_future_annotation.parallelized
    assert node.Node.from_fn(parallelized).node_role == node.NodeType.EXPAND


def test_node_from_future_annotation_standard():
    standard = nodes_with_future_annotation.standard
    assert node.Node.from_fn(standard).node_role == node.NodeType.STANDARD


def test_node_from_future_annotation_collected():
    collected = nodes_with_future_annotation.collected
    assert node.Node.from_fn(collected).node_role == node.NodeType.COLLECT



import dataclasses
from collections import Counter
from typing import Any, Collection, Dict, List, Tuple, Type

import pandas as pd
import pytest

from hamilton import ad_hoc_utils, base, driver, graph, node
from hamilton.function_modifiers import base as fm_base
from hamilton.function_modifiers import extract_fields, save_to, source, value
from hamilton.function_modifiers.adapters import (
    InvalidDecoratorException,
    LoadFromDecorator,
    SaveToDecorator,
    dataloader,
    datasaver,
    load_from,
    resolve_adapter_class,
    resolve_kwargs,
)
from hamilton.function_modifiers.base import DefaultNodeCreator
from hamilton.io.data_adapters import DataLoader, DataSaver
from hamilton.io.default_data_loaders import JSONDataSaver
from hamilton.registry import LOADER_REGISTRY


def test_default_adapters_are_available():
    assert len(LOADER_REGISTRY) > 0


def test_default_adapters_are_registered_once():
    assert "json" in LOADER_REGISTRY
    count_unique = {
        # we want str() of the class to get the fully qualified class name.
        key: Counter([str(value) for value in values])
        for key, values in LOADER_REGISTRY.items()
    }
    for key, value_ in count_unique.items():
        for impl, count in value_.items():
            assert count == 1, (
                f"Adapter for {key} registered multiple times for {impl}. This should not"
                f" happen, as items should just be registered once."
            )


@dataclasses.dataclass
class MockDataLoader(DataLoader):
    required_param: int
    required_param_2: int
    required_param_3: str
    default_param: int = 4
    default_param_2: int = 5
    default_param_3: str = "6"

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [int]

    def load_data(self, type_: Type[int]) -> Tuple[int, Dict[str, Any]]:
        return ..., {"required_param": self.required_param, "default_param": self.default_param}

    @classmethod
    def name(cls) -> str:
        return "mock"


def test_load_from_decorator():
    def fn(data: int) -> int:
        return data

    decorator = LoadFromDecorator(
        [MockDataLoader],
        required_param=value("1"),
        required_param_2=value("2"),
        required_param_3=value("3"),
    )
    nodes_raw = DefaultNodeCreator().generate_nodes(fn, {})
    nodes = decorator.transform_dag(nodes_raw, {}, fn)
    assert len(nodes) == 3
    nodes_by_name = {node_.name: node_ for node_ in nodes}
    assert len(nodes_by_name) == 3
    assert "fn" in nodes_by_name
    assert nodes_by_name["fn.load_data.data"].tags == {
        "hamilton.data_loader.source": "mock",
        "hamilton.data_loader": True,
        "hamilton.data_loader.has_metadata": True,
        "hamilton.data_loader.node": "data",
        "hamilton.data_loader.classname": MockDataLoader.__qualname__,
    }
    assert nodes_by_name["fn.select_data.data"].tags == {
        "hamilton.data_loader.source": "mock",
        "hamilton.data_loader": True,
        "hamilton.data_loader.has_metadata": False,
        "hamilton.data_loader.node": "data",
        "hamilton.data_loader.classname": MockDataLoader.__qualname__,
    }


def test_load_from_decorator_resolve_kwargs():
    kwargs = dict(
        required_param=source("1"),
        required_param_2=value(2),
        required_param_3=value("3"),
        default_param=source("4"),
        default_param_2=value(5),
    )

    dependency_kwargs, literal_kwargs = resolve_kwargs(kwargs)
    assert dependency_kwargs == {"required_param": "1", "default_param": "4"}
    assert literal_kwargs == {"required_param_2": 2, "required_param_3": "3", "default_param_2": 5}


def test_load_from_decorator_resolve_kwargs_with_literals():
    kwargs = dict(
        required_param=source("1"),
        required_param_2=2,
        required_param_3="3",
        default_param=source("4"),
        default_param_2=5,
    )

    dependency_kwargs, literal_kwargs = resolve_kwargs(kwargs)
    assert dependency_kwargs == {"required_param": "1", "default_param": "4"}
    assert literal_kwargs == {"required_param_2": 2, "required_param_3": "3", "default_param_2": 5}


def test_load_from_decorator_validate_succeeds():
    decorator = LoadFromDecorator(
        [MockDataLoader],
        required_param=source("1"),
        required_param_2=value(2),
        required_param_3=value("3"),
    )

    def fn(injected_data: int) -> int:
        return injected_data

    decorator.validate(fn)


def test_load_from_decorator_validate_succeeds_with_inject():
    decorator = LoadFromDecorator(
        [MockDataLoader],
        inject_="injected_data",
        required_param=source("1"),
        required_param_2=value(2),
        required_param_3=value("3"),
    )

    def fn(injected_data: int, dependent_data: int) -> int:
        return injected_data + dependent_data

    decorator.validate(fn)


def test_load_from_decorator_validate_fails_dont_know_which_param_to_inject():
    decorator = LoadFromDecorator(
        [MockDataLoader],
        required_param=source("1"),
        required_param_2=value(2),
        required_param_3=value("3"),
    )

    def fn(injected_data: int, other_possible_injected_data: int) -> int:
        return injected_data + other_possible_injected_data

    with pytest.raises(fm_base.InvalidDecoratorException):
        decorator.validate(fn)


def test_load_from_decorator_validate_fails_inject_not_in_fn():
    decorator = LoadFromDecorator(
        [MockDataLoader],
        inject_="injected_data",
        required_param=source("1"),
        required_param_2=value(2),
        required_param_3=value("3"),
    )

    def fn(dependent_data: int) -> int:
        return dependent_data

    with pytest.raises(fm_base.InvalidDecoratorException):
        decorator.validate(fn)


def test_load_from_decorator_validate_fails_inject_missing_param():
    decorator = LoadFromDecorator(
        [MockDataLoader],
        required_param=source("1"),
        required_param_2=value(2),
        # This is commented out cause it'll be missing
        # required_param_3=value("3"),
    )

    def fn(data: int) -> int:
        return data

    with pytest.raises(fm_base.InvalidDecoratorException):
        decorator.validate(fn)


@dataclasses.dataclass
class StringDataLoader(DataLoader):
    def load_data(self, type_: Type) -> Tuple[str, Dict[str, Any]]:
        return "foo", {"loader": "string_data_loader"}

    @classmethod
    def name(cls) -> str:
        return "dummy"

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [str]


@dataclasses.dataclass
class AnyDataLoader(DataLoader):
    value: Any

    def load_data(self, type_: Type) -> Tuple[Any, Dict[str, Any]]:
        return self.value, {"loader": "any_data_loader"}

    @classmethod
    def name(cls) -> str:
        return "dummy"

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [Any]


@dataclasses.dataclass
class IntDataLoader(DataLoader):
    def load_data(self, type_: Type) -> Tuple[int, Dict[str, Any]]:
        return 1, {"loader": "int_data_loader"}

    @classmethod
    def name(cls) -> str:
        return "dummy"

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [int]


@dataclasses.dataclass
class IntDataLoader2(DataLoader):
    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [int]

    def load_data(self, type_: Type) -> Tuple[int, Dict[str, Any]]:
        return 2, {"loader": "int_data_loader_2"}

    @classmethod
    def name(cls) -> str:
        return "dummy"


def test_validate_fails_incorrect_type():
    decorator = LoadFromDecorator(
        [StringDataLoader, IntDataLoader],
    )

    def fn_str_inject(injected_data: str) -> str:
        return injected_data

    def fn_int_inject(injected_data: int) -> int:
        return injected_data

    def fn_bool_inject(injected_data: bool) -> bool:
        return injected_data

    # This is valid as there is one parameter and its a type that the decorator supports
    decorator.validate(fn_str_inject)

    # This is valid as there is one parameter and its a type that the decorator supports
    decorator.validate(fn_int_inject)

    # This is invalid as there is one parameter and it is not a type that the decorator supports
    with pytest.raises(fm_base.InvalidDecoratorException):
        decorator.validate(fn_bool_inject)


def test_validate_selects_correct_type():
    decorator = LoadFromDecorator(
        [StringDataLoader, IntDataLoader],
    )

    def fn_str_inject(injected_data: str) -> str:
        return injected_data

    def fn_int_inject(injected_data: int) -> int:
        return injected_data

    def fn_bool_inject(injected_data: bool) -> bool:
        return injected_data

    # This is valid as there is one parameter and its a type that the decorator supports
    decorator.validate(fn_str_inject)

    # This is valid as there is one parameter and its a type that the decorator supports
    decorator.validate(fn_int_inject)

    # This is invalid as there is one parameter and it is not a type that the decorator supports
    with pytest.raises(fm_base.InvalidDecoratorException):
        decorator.validate(fn_bool_inject)


# Note that this tests an internal API, but we would like to test this to ensure
# class selection is correct
@pytest.mark.parametrize(
    "type_,classes,correct_class",
    [
        (str, [StringDataLoader, IntDataLoader, IntDataLoader2], StringDataLoader),
        (int, [StringDataLoader, IntDataLoader, IntDataLoader2], IntDataLoader2),
        (int, [IntDataLoader2, IntDataLoader], IntDataLoader),
        (int, [IntDataLoader, IntDataLoader2], IntDataLoader2),
        (int, [StringDataLoader], None),
        (str, [IntDataLoader], None),
        (dict, [IntDataLoader], None),
        (dict, [IntDataLoader, StringDataLoader], None),
        (str, [AnyDataLoader, StringDataLoader], StringDataLoader),
        (dict, [AnyDataLoader, StringDataLoader], AnyDataLoader),
    ],
)
def test_resolve_correct_loader_class(
    type_: Type[Type], classes: List[Type[DataLoader]], correct_class: Type[DataLoader]
):
    assert resolve_adapter_class(type_, classes) == correct_class


def test_decorator_validate():
    decorator = LoadFromDecorator(
        [StringDataLoader, IntDataLoader, IntDataLoader2],
    )

    def fn_str_inject(injected_data: str) -> str:
        return injected_data

    def fn_int_inject(injected_data: int) -> int:
        return injected_data

    def fn_bool_inject(injected_data: bool) -> bool:
        return injected_data

    # This is valid as there is one parameter and its a type that the decorator supports
    decorator.validate(fn_str_inject)
    decorator.validate(fn_int_inject)
    # This is invalid as there is one parameter and it is not a type that the decorator supports
    with pytest.raises(fm_base.InvalidDecoratorException):
        decorator.validate(fn_bool_inject)


# End-to-end tests are probably cleanest
# We've done a bunch of tests of internal structures for other decorators,
# but that leaves the testing brittle
# We don't test the driver, we just use the function_graph to tests the nodes
def test_load_from_decorator_end_to_end():
    @LoadFromDecorator(
        [StringDataLoader, IntDataLoader, IntDataLoader2],
    )
    def fn_str_inject(injected_data: str) -> str:
        return injected_data

    fg = graph.FunctionGraph.from_modules(
        ad_hoc_utils.create_temporary_module(fn_str_inject), config={}
    )
    result = fg.execute(inputs={}, nodes=fg.nodes.values())
    assert result["fn_str_inject"] == "foo"
    assert result["fn_str_inject.load_data.injected_data"] == (
        "foo",
        {"loader": "string_data_loader"},
    )


# End-to-end tests are probably cleanest
# We've done a bunch of tests of internal structures for other decorators,
# but that leaves the testing brittle
# We don't test the driver, we just use the function_graph to tests the nodes
def test_load_from_decorator_end_to_end_with_multiple():
    @LoadFromDecorator(
        [StringDataLoader, IntDataLoader, IntDataLoader2],
        inject_="injected_data_1",
    )
    @LoadFromDecorator(
        [StringDataLoader, IntDataLoader, IntDataLoader2],
        inject_="injected_data_2",
    )
    def fn_str_inject(injected_data_1: str, injected_data_2: int) -> str:
        return "".join([injected_data_1] * injected_data_2)

    fg = graph.FunctionGraph.from_modules(
        ad_hoc_utils.create_temporary_module(fn_str_inject), config={}
    )
    result = fg.execute(inputs={}, nodes=fg.nodes.values())
    assert result["fn_str_inject"] == "foofoo"
    assert result["fn_str_inject.load_data.injected_data_1"] == (
        "foo",
        {"loader": "string_data_loader"},
    )

    assert result["fn_str_inject.load_data.injected_data_2"] == (
        2,
        {"loader": "int_data_loader_2"},
    )


@pytest.mark.parametrize(
    "source_",
    [
        value("tests/resources/data/test_load_from_data.json"),
        source("test_data"),
    ],
)
def test_load_from_decorator_json_file(source_):
    @load_from.json(path=source_)
    def raw_json_data(data: Dict[str, Any]) -> Dict[str, Any]:
        return data

    def number_employees(raw_json_data: Dict[str, Any]) -> int:
        return len(raw_json_data["employees"])

    def sum_age(raw_json_data: Dict[str, Any]) -> float:
        return sum([employee["age"] for employee in raw_json_data["employees"]])

    def mean_age(sum_age: float, number_employees: int) -> float:
        return sum_age / number_employees

    config = {}
    dr = driver.Driver(
        config,
        ad_hoc_utils.create_temporary_module(raw_json_data, number_employees, sum_age, mean_age),
        adapter=base.DefaultAdapter(),
    )
    result = dr.execute(
        ["mean_age"], inputs={"test_data": "tests/resources/data/test_load_from_data.json"}
    )
    assert result["mean_age"] - 32.33333 < 0.0001


def test_loader_fails_for_missing_attribute():
    with pytest.raises(AttributeError):
        load_from.not_a_loader(param=value("foo"))


def test_pandas_extensions_end_to_end(tmp_path_factory):
    output_path = str(tmp_path_factory.mktemp("test_pandas_extensions_end_to_end") / "output.csv")
    input_path = "tests/resources/data/test_load_from_data.csv"

    @save_to.csv(path=source("output_path"), output_name_="save_df")
    @load_from.csv(path=source("input_path"))
    def df(data: pd.DataFrame) -> pd.DataFrame:
        return data

    config = {}
    dr = driver.Driver(
        config,
        ad_hoc_utils.create_temporary_module(df),
        adapter=base.DefaultAdapter(),
    )
    # run once to check that loading is correct
    result = dr.execute(
        ["df", "save_df"],
        inputs={"input_path": input_path, "output_path": output_path},
    )
    assert result["df"].shape == (3, 5)
    assert result["df"].loc[0, "firstName"] == "John"

    #
    result_just_read = dr.execute(
        ["df"],
        inputs={"input_path": output_path},
    )
    # This is just reading the same file we wrote out, so it should be the same
    pd.testing.assert_frame_equal(result["df"], result_just_read["df"])


@dataclasses.dataclass
class MarkingSaver(DataSaver):
    markers: set
    more_markers: set

    def save_data(self, data: int) -> Dict[str, Any]:
        self.markers.add(data)
        self.more_markers.add(data)
        return {}

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [int, dict]

    @classmethod
    def name(cls) -> str:
        return "marker"


def test_save_to_decorator():
    def fn() -> int:
        return 1

    marking_set = set()
    marking_set_2 = set()
    decorator = SaveToDecorator(
        [MarkingSaver],
        output_name_="save_fn",
        markers=value(marking_set),
        more_markers=source("more_markers"),
    )
    node_ = node.Node.from_fn(fn)
    nodes = decorator.transform_node(node_, {}, fn)
    assert len(nodes) == 2
    nodes_by_name = {node_.name: node_ for node_ in nodes}
    assert "save_fn" in nodes_by_name
    assert "fn" in nodes_by_name
    save_fn_node = nodes_by_name["save_fn"]
    assert sorted(save_fn_node.input_types.keys()) == ["fn", "more_markers"]
    assert save_fn_node(**{"fn": 1, "more_markers": marking_set_2}) == {}
    assert save_fn_node.tags == {
        "hamilton.data_saver": True,
        "hamilton.data_saver.sink": "marker",
        "hamilton.data_saver.classname": MarkingSaver.__qualname__,
    }
    # Check that the markers are updated, ensuring that the save_fn is called
    assert marking_set_2 == {1}
    assert marking_set == {1}


def test_save_to_decorator_with_target():
    @extract_fields({"a": int, "b": int})
    def fn() -> dict:
        return {"a": 1, "b": 2}

    decorator = SaveToDecorator(
        [JSONDataSaver],
        output_name_="save_fn",
        path=value("unused"),
        target_="fn",  # save the original
    )
    node_ = node.Node.from_fn(fn)
    nodes = {n.name: n for n in decorator.transform_node(node_, {}, fn)}
    saver_node = nodes["save_fn"]
    # We just want to make sure it gets the right input
    assert list(saver_node.input_types) == ["fn"]


@dataclasses.dataclass
class DefaultFactoryLoader(DataLoader):
    field_with_factory: int = dataclasses.field(default_factory=int)

    def __post_init__(self):
        self.param2 = self.field_with_factory + 1

    def load_data(self, type_: Type[int]) -> Tuple[int, Dict[str, Any]]:
        return self.param2, {}

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [int]

    @classmethod
    def name(cls) -> str:
        return "factory"


def test_loader_default_factory_field():
    @LoadFromDecorator([DefaultFactoryLoader])
    def foo(param: int) -> int:
        return param

    fg = graph.create_function_graph(
        ad_hoc_utils.create_temporary_module(foo),
        config={},
    )
    assert len(fg) == 3
    assert "foo" in fg


@dataclasses.dataclass
class DefaultFactorySaver(DataSaver):
    field_with_factory: int = dataclasses.field(default_factory=int)

    def __post_init__(self):
        self.param2 = self.field_with_factory + 1

    def save_data(self, data: int) -> Dict[str, Any]:
        return {}

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [int]

    @classmethod
    def name(cls) -> str:
        return "factory"


def test_saver_default_factory_field():
    @SaveToDecorator([DefaultFactorySaver])
    def foo(param: int) -> int:
        return param

    fg = graph.create_function_graph(
        ad_hoc_utils.create_temporary_module(foo),
        config={},
    )
    assert len(fg) == 3
    assert "foo" in fg


@dataclasses.dataclass
class OptionalParamDataLoader(DataLoader):
    param: int = 1

    def load_data(self, type_: Type[int]) -> Tuple[int, Dict[str, Any]]:
        return self.param, {}

    @classmethod
    def applicable_types(cls) -> Collection[Type]:
        return [int]

    @classmethod
    def name(cls) -> str:
        return "optional"


def test_adapters_optional_params():
    @LoadFromDecorator([OptionalParamDataLoader])
    def foo(param: int) -> int:
        return param

    fg = graph.create_function_graph(
        ad_hoc_utils.create_temporary_module(foo),
        config={},
    )
    assert len(fg) == 3
    assert "foo" in fg


def test_save_to_with_input_from_other_fn():
    # This tests that we can refer to another node in save_to
    def output_path() -> str:
        return "output.json"

    @save_to.json(path=source("output_path"), output_name_="save_fn")
    def fn() -> dict:
        return {"a": 1}

    fg = graph.create_function_graph(
        ad_hoc_utils.create_temporary_module(output_path, fn),
        config={},
    )

    assert len(fg) == 3


def test_load_from_with_input_from_other_fn():
    # This tests that we can refer to another node in load_from
    def input_path() -> str:
        return "input.json"

    @load_from.json(path=source("input_path"))
    def fn(data: dict) -> dict:
        return data

    fg = graph.create_function_graph(
        ad_hoc_utils.create_temporary_module(input_path, fn),
        config={},
    )
    assert len(fg) == 4


def test_load_from_with_multiple_inputs():
    # This tests that we can refer to another node in load_from

    @load_from.json(
        path=value("input_1.json"),
        inject_="data1",
    )
    @load_from.json(
        path=value("input_2.json"),
        inject_="data2",
    )
    def fn(data1: dict, data2: dict) -> dict:
        return {**data1, **data2}

    fg = graph.create_function_graph(
        ad_hoc_utils.create_temporary_module(fn),
        config={},
    )
    # One filter, one loader for each and the transform function
    assert len(fg) == 5


import sys

if sys.version_info >= (3, 9):
    dl_type = tuple[int, dict]
    ds_type = dict
else:
    dl_type = Tuple[int, Dict]
    ds_type = Dict


# Mock functions for dataloader & datasaver testing
def correct_dl_function(foo: int) -> dl_type:
    return 1, {}


def correct_ds_function(data: float) -> ds_type:
    return {}


def no_return_annotation_function():
    return 1, {}


def non_tuple_return_function() -> int:
    return 1


def incorrect_tuple_length_function() -> Tuple[int]:
    return (1,)


def incorrect_second_element_function() -> Tuple[int, list]:
    return 1, []


incorrect_funcs = [
    no_return_annotation_function,
    non_tuple_return_function,
    incorrect_tuple_length_function,
    incorrect_second_element_function,
]


@pytest.mark.parametrize("func", incorrect_funcs, ids=[f.__name__ for f in incorrect_funcs])
def test_dl_validate_incorrect_functions(func):
    dl = dataloader()
    with pytest.raises(InvalidDecoratorException):
        dl.validate(func)


def test_dl_validate_with_correct_function():
    dl = dataloader()
    try:
        dl.validate(correct_dl_function)
    except InvalidDecoratorException:
        # i.e. fail the test if there's an error
        pytest.fail("validate() raised InvalidDecoratorException unexpectedly!")


def test_ds_validate_with_correct_function():
    dl = datasaver()
    try:
        dl.validate(correct_ds_function)
    except InvalidDecoratorException:
        # i.e. fail the test if there's an error
        pytest.fail("validate() raised InvalidDecoratorException unexpectedly!")


def test_ds_validate_incorrect_function():
    dl = dataloader()
    with pytest.raises(InvalidDecoratorException):
        dl.validate(non_tuple_return_function)


def test_dataloader():
    annotation = dataloader()
    (node1, node2) = annotation.generate_nodes(correct_dl_function, {})
    assert node1.name == "correct_dl_function.loader"
    assert node1.input_types["foo"][1] == node.DependencyType.REQUIRED
    assert node1.callable(foo=0) == (1, {})
    assert node1.tags == {
        "hamilton.data_loader": True,
        "hamilton.data_loader.classname": "correct_dl_function()",
        "hamilton.data_loader.has_metadata": True,
        "hamilton.data_loader.node": "correct_dl_function",
        "hamilton.data_loader.source": "loader",
        "module": "tests.function_modifiers.test_adapters",
    }
    assert node2.name == "correct_dl_function"
    assert node2.callable(**{"correct_dl_function.loader": (1, {})}) == 1
    assert node2.tags == {
        "hamilton.data_loader": True,
        "hamilton.data_loader.classname": "correct_dl_function()",
        "hamilton.data_loader.has_metadata": False,
        "hamilton.data_loader.node": "correct_dl_function",
        "hamilton.data_loader.source": "correct_dl_function",
    }


def test_datasaver():
    annotation = datasaver()
    (node1,) = annotation.generate_nodes(correct_ds_function, {})
    assert node1.name == "correct_ds_function"
    assert node1.input_types["data"][1] == node.DependencyType.REQUIRED
    assert node1.callable(data=0.0) == {}
    assert node1.tags == {
        "hamilton.data_saver": True,
        "hamilton.data_saver.classname": "correct_ds_function()",
        "hamilton.data_saver.sink": "correct_ds_function",
        "module": "tests.function_modifiers.test_adapters",
    }



from typing import Any, Type

import pyarrow
from pyarrow.interchange import from_dataframe

from hamilton.lifecycle.api import ResultBuilder


class PyarrowTableResult(ResultBuilder):
    """Add this result builder to a materializer's `combine` statement to convert your dataframe
    object to a pyarrow representation and make it compatible with pyarrow DataSavers.

    It implicitly support input_type == Any, but it expects dataframe objects implementing
    the dataframe interchange protocol: ref: https://arrow.apache.org/docs/python/interchange_protocol.html
    for example:
    - pandas
    - polars
    - dask
    - vaex
    - ibis
    - duckdb results
    """

    def output_type(self) -> Type:
        return pyarrow.Table

    def build_result(self, **outputs: Any) -> Any:
        """This function converts objects implementing the `__dataframe__` protocol to
        a pyarrow table. It doesn't support receiving multiple outputs because it can't
        handle any joining logic.

        ref: https://arrow.apache.org/docs/python/interchange_protocol.html
        """
        if len(outputs) != 1:
            raise AssertionError(
                "PyarrowTableResult can only receive 1 output, i.e., only one item in `to.SAVER(dependencies=[])`"
                f"It received {len(outputs)} outputs."
            )
        return from_dataframe(next(iter(outputs.values())))



import inspect
from typing import List, Set

import pandas as pd
import pytest

import hamilton.function_modifiers
from hamilton import base, driver, function_modifiers, models, node
from hamilton.function_modifiers import does
from hamilton.function_modifiers.dependencies import source, value
from hamilton.function_modifiers.macros import (
    Applicable,
    apply_to,
    ensure_function_empty,
    mutate,
    pipe_input,
    pipe_output,
    step,
)
from hamilton.node import DependencyType

import tests.resources.mutate
import tests.resources.pipe_input
import tests.resources.pipe_output


def test_no_code_validator():
    def no_code():
        pass

    def no_code_with_docstring():
        """This should still show up as having no code, even though it has a docstring"""
        pass

    def yes_code():
        """This should show up as having no code"""
        a = 0
        return a

    ensure_function_empty(no_code)
    ensure_function_empty(no_code_with_docstring)
    with pytest.raises(hamilton.function_modifiers.base.InvalidDecoratorException):
        ensure_function_empty(yes_code)


# Functions for  @does -- these are the functions we're "replacing"
def _no_params() -> int:
    pass


def _one_param(a: int) -> int:
    pass


def _two_params(a: int, b: int) -> int:
    pass


def _three_params(a: int, b: int, c: int) -> int:
    pass


def _three_params_with_defaults(a: int, b: int = 1, c: int = 2) -> int:
    pass


def _empty() -> int:
    return 1


def _kwargs(**kwargs: int) -> int:
    return sum(kwargs.values())


def _kwargs_with_a(a: int, **kwargs: int) -> int:
    return a + sum(kwargs.values())


def _just_a(a: int) -> int:
    return a


def _just_b(b: int) -> int:
    return b


def _a_b_c(a: int, b: int, c: int) -> int:
    return a + b + c


@pytest.mark.parametrize(
    "fn,replace_with,argument_mapping,matches",
    [
        (_no_params, _empty, {}, True),
        (_no_params, _kwargs, {}, True),
        (_no_params, _kwargs_with_a, {}, False),
        (_no_params, _just_a, {}, False),
        (_no_params, _a_b_c, {}, False),
        (_one_param, _empty, {}, False),
        (_one_param, _kwargs, {}, True),
        (_one_param, _kwargs_with_a, {}, True),
        (_one_param, _just_a, {}, True),
        (_one_param, _just_b, {}, False),
        (_one_param, _just_b, {"b": "a"}, True),  # Replacing a with b makes the signatures match
        (_one_param, _just_b, {"c": "a"}, False),  # Replacing a with b makes the signatures match
        (_two_params, _empty, {}, False),
        (_two_params, _kwargs, {}, True),
        (_two_params, _kwargs_with_a, {}, True),  # b gets fed to kwargs
        (_two_params, _kwargs_with_a, {"foo": "b"}, True),  # Any kwargs work
        (_two_params, _kwargs_with_a, {"bar": "a"}, False),  # No param bar
        (_two_params, _just_a, {}, False),
        (_two_params, _just_b, {}, False),
        (_three_params, _a_b_c, {}, True),
        (_three_params, _a_b_c, {"d": "a"}, False),
        (_three_params, _a_b_c, {}, True),
        (_three_params, _a_b_c, {"a": "b", "b": "a"}, True),  # Weird case but why not?
        (_three_params, _kwargs_with_a, {}, True),
        (_three_params_with_defaults, _a_b_c, {}, True),
        (_three_params_with_defaults, _a_b_c, {"d": "a"}, False),
        (_three_params_with_defaults, _a_b_c, {}, True),
    ],
)
def test_ensure_function_signatures_compatible(fn, replace_with, argument_mapping, matches):
    assert (
        does.test_function_signatures_compatible(
            inspect.signature(fn), inspect.signature(replace_with), argument_mapping
        )
        == matches
    )


def test_does_function_modifier():
    def sum_(**kwargs: int) -> int:
        return sum(kwargs.values())

    def to_modify(param1: int, param2: int) -> int:
        """This sums the inputs it gets..."""
        pass

    annotation = does(sum_)
    (node_,) = annotation.generate_nodes(to_modify, {})
    assert node_.name == "to_modify"
    assert node_.callable(param1=1, param2=1) == 2
    assert node_.documentation == to_modify.__doc__


def test_does_function_modifier_complex_types():
    def setify(**kwargs: List[int]) -> Set[int]:
        return set(sum(kwargs.values(), []))

    def to_modify(param1: List[int], param2: List[int]) -> int:
        """This sums the inputs it gets..."""
        pass

    annotation = does(setify)
    (node_,) = annotation.generate_nodes(to_modify, {})
    assert node_.name == "to_modify"
    assert node_.callable(param1=[1, 2, 3], param2=[4, 5, 6]) == {1, 2, 3, 4, 5, 6}
    assert node_.documentation == to_modify.__doc__


def test_does_function_modifier_optionals():
    def sum_(param0: int, **kwargs: int) -> int:
        return sum(kwargs.values())

    def to_modify(param0: int, param1: int = 1, param2: int = 2) -> int:
        """This sums the inputs it gets..."""
        pass

    annotation = does(sum_)
    (node_,) = annotation.generate_nodes(to_modify, {})
    assert node_.name == "to_modify"
    assert node_.input_types["param0"][1] == DependencyType.REQUIRED
    assert node_.input_types["param1"][1] == DependencyType.OPTIONAL
    assert node_.input_types["param2"][1] == DependencyType.OPTIONAL
    assert node_.callable(param0=0) == 3
    assert node_.callable(param0=0, param1=0, param2=0) == 0
    assert node_.documentation == to_modify.__doc__


def test_does_with_argument_mapping():
    def _sum_multiply(param0: int, param1: int, param2: int) -> int:
        return param0 + param1 * param2

    def to_modify(parama: int, paramb: int = 1, paramc: int = 2) -> int:
        """This sums the inputs it gets..."""
        pass

    annotation = does(_sum_multiply, param0="parama", param1="paramb", param2="paramc")
    (node_,) = annotation.generate_nodes(to_modify, {})
    assert node_.name == "to_modify"
    assert node_.input_types["parama"][1] == DependencyType.REQUIRED
    assert node_.input_types["paramb"][1] == DependencyType.OPTIONAL
    assert node_.input_types["paramc"][1] == DependencyType.OPTIONAL
    assert node_.callable(parama=0) == 2
    assert node_.callable(parama=0, paramb=1, paramc=2) == 2
    assert node_.callable(parama=1, paramb=4) == 9
    assert node_.documentation == to_modify.__doc__


def test_model_modifier():
    config = {
        "my_column_model_params": {
            "col_1": 0.5,
            "col_2": 0.5,
        }
    }

    class LinearCombination(models.BaseModel):
        def get_dependents(self) -> List[str]:
            return list(self.config_parameters.keys())

        def predict(self, **columns: pd.Series) -> pd.Series:
            return sum(
                self.config_parameters[column_name] * column
                for column_name, column in columns.items()
            )

    def my_column() -> pd.Series:
        """Column that will be annotated by a model"""
        pass

    annotation = function_modifiers.model(LinearCombination, "my_column_model_params")
    annotation.validate(my_column)
    (model_node,) = annotation.generate_nodes(my_column, config)
    assert model_node.input_types["col_1"][0] == model_node.input_types["col_2"][0] == pd.Series
    assert model_node.type == pd.Series
    pd.testing.assert_series_equal(
        model_node.callable(col_1=pd.Series([1]), col_2=pd.Series([2])), pd.Series([1.5])
    )

    def bad_model(col_1: pd.Series, col_2: pd.Series) -> pd.Series:
        return col_1 * 0.5 + col_2 * 0.5

    with pytest.raises(hamilton.function_modifiers.base.InvalidDecoratorException):
        annotation.validate(bad_model)


def _test_apply_function(foo: int, bar: int, baz: int = 100) -> int:
    return foo + bar + baz


@pytest.mark.parametrize(
    "args,kwargs,chain_first_param",
    [
        ([source("foo_upstream"), value(1)], {}, False),
        ([value(1)], {}, True),
        ([source("foo_upstream"), value(1), value(2)], {}, False),
        ([value(1), value(2)], {}, True),
        ([source("foo_upstream")], {"bar": value(1)}, False),
        ([], {"bar": value(1)}, True),
        ([], {"foo": source("foo_upstream"), "bar": value(1)}, False),
        ([], {"bar": value(1)}, True),
        ([], {"foo": source("foo_upstream"), "bar": value(1), "baz": value(1)}, False),
        ([], {"bar": value(1), "baz": value(1)}, True),
    ],
)
def test_applicable_validates_correctly(args, kwargs, chain_first_param: bool):
    applicable = Applicable(_test_apply_function, args=args, kwargs=kwargs)
    applicable.validate(chain_first_param=chain_first_param, allow_custom_namespace=True)


@pytest.mark.parametrize(
    "args,kwargs,chain_first_param",
    [
        (
            [source("foo_upstream"), value(1)],
            {"foo": source("foo_upstream")},
            True,
        ),  # We chain the first parameter, not pass it in
        ([value(1)], {}, False),  # Not enough first parameters
        ([source("foo_upstream"), value(1), value(2)], {}, True),
        ([value(2)], {"foo": source("foo_upstream")}, False),
        ([source("foo_upstream")], {"bar": value(1)}, True),
        ([], {"bar": value(1)}, False),
        ([], {"foo": source("foo_upstream"), "bar": value(1)}, True),
        ([], {"bar": value(1)}, False),
        ([], {"foo": source("foo_upstream"), "bar": value(1), "baz": value(1)}, True),
        ([], {"bar": value(1), "baz": value(1)}, False),
    ],
)
def test_applicable_does_not_validate(args, kwargs, chain_first_param: bool):
    with pytest.raises(hamilton.function_modifiers.base.InvalidDecoratorException):
        applicable = Applicable(_test_apply_function, args=args, kwargs=kwargs)
        applicable.validate(chain_first_param=chain_first_param, allow_custom_namespace=True)


def test_applicable_does_not_validate_invalid_function_pos_only():
    def foo(a: int, /, b: int) -> int:
        return a + b

    with pytest.raises(hamilton.function_modifiers.base.InvalidDecoratorException):
        applicable = Applicable(foo, args=(source("a"), source("b")), kwargs={})
        applicable.validate(chain_first_param=True, allow_custom_namespace=True)


# We will likely start supporting this in the future, but for now we don't
def test_applicable_does_not_validate_no_param_type_hints():
    def foo(a, b) -> int:
        return a + b

    with pytest.raises(hamilton.function_modifiers.base.InvalidDecoratorException):
        applicable = Applicable(foo, args=(source("a"), source("b")), kwargs={})
        applicable.validate(chain_first_param=True, allow_custom_namespace=True)


def test_applicable_does_not_validate_no_return_type_hints():
    def foo(a: int, b: int):
        return a + b

    with pytest.raises(hamilton.function_modifiers.base.InvalidDecoratorException):
        applicable = Applicable(foo, args=(source("a"), source("b")), kwargs={})
        applicable.validate(chain_first_param=True, allow_custom_namespace=True)


def test_applicable_does_not_validate_invalid_function_no_params():
    def foo() -> int:
        return 1

    with pytest.raises(hamilton.function_modifiers.base.InvalidDecoratorException):
        applicable = Applicable(foo, args=(), kwargs={})
        applicable.validate(chain_first_param=True, allow_custom_namespace=True)


def general_downstream_function(result: int) -> int:
    return result


def function_multiple_same_type_params(p1: int, p2: int, p3: int) -> int:
    return p1 + p2 + p3


# TODO: in case of multiple paramters need some type checking
# def function_multiple_diverse_type_params(p1: int, p2: str, p3: int) -> int:
#     return p1 + len(p2) + p3


def test_pipe_input_on_input_error_unless_string_or_none():
    with pytest.raises(NotImplementedError):
        decorator = pipe_input(  # noqa
            step(_test_apply_function, source("bar_upstream"), baz=value(10)).named("node_1"),
            step(_test_apply_function, source("bar_upstream"), baz=value(100)).named("node_2"),
            step(_test_apply_function, source("bar_upstream"), baz=value(1000)).named("node_3"),
            on_input=["p2", "p3"],
            namespace="abc",
        )


def test_pipe_input_mapping_args_targets_global():
    n = node.Node.from_fn(function_multiple_same_type_params)

    decorator = pipe_input(
        step(_test_apply_function, source("bar_upstream"), baz=value(10)).named("node_1"),
        step(_test_apply_function, source("bar_upstream"), baz=value(100)).named("node_2"),
        step(_test_apply_function, source("bar_upstream"), baz=value(1000)).named("node_3"),
        on_input="p2",
        namespace="abc",
    )
    nodes = decorator.transform_dag([n], {}, function_multiple_same_type_params)
    nodes_by_name = {item.name: item for item in nodes}
    chain_node = nodes_by_name["abc.node_1"]
    assert chain_node(p2=1, bar_upstream=3) == 14


# TODO: multiple parameter tests
# def test_pipe_input_no_namespace_with_target():
#     n = node.Node.from_fn(function_multiple_diverse_type_params)

#     decorator = pipe_input(
#         step(_test_apply_function, source("bar_upstream"), baz=value(10))
#         .on_input(["p1", "p3"])
#         .named("node_1"),
#         step(_test_apply_function, source("bar_upstream"), baz=value(100)).named("node_2"),
#         step(_test_apply_function, source("bar_upstream"), baz=value(1000))
#         .on_input("p3")
#         .named("node_3"),
#         on_input="p2",
#         namespace=None,
#     )
#     nodes = decorator.transform_dag([n], {}, function_multiple_diverse_type_params)
#     final_node = nodes[0].name
#     p1_node = nodes[1].name
#     p2_node1 = nodes[2].name
#     p2_node2 = nodes[3].name
#     p2_node3 = nodes[4].name
#     p3_node1 = nodes[5].name
#     p3_node2 = nodes[6].name

#     assert final_node == "function_multiple_diverse_type_params"
#     assert p1_node == "p1.node_1"
#     assert p2_node1 == "p2.node_1"
#     assert p2_node2 == "p2.node_2"
#     assert p2_node3 == "p2.node_3"
#     assert p3_node1 == "p3.node_1"
#     assert p3_node2 == "p3.node_3"


# def test_pipe_input_elipsis_namespace_with_target():
#     n = node.Node.from_fn(function_multiple_diverse_type_params)

#     decorator = pipe_input(
#         step(_test_apply_function, source("bar_upstream"), baz=value(10))
#         .on_input(["p1", "p3"])
#         .named("node_1"),
#         step(_test_apply_function, source("bar_upstream"), baz=value(100)).named("node_2"),
#         step(_test_apply_function, source("bar_upstream"), baz=value(1000))
#         .on_input("p3")
#         .named("node_3"),
#         namespace=...,
#         on_input="p2",
#     )
#     nodes = decorator.transform_dag([n], {}, function_multiple_diverse_type_params)
#     final_node = nodes[0].name
#     p1_node = nodes[1].name
#     p2_node1 = nodes[2].name
#     p2_node2 = nodes[3].name
#     p2_node3 = nodes[4].name
#     p3_node1 = nodes[5].name
#     p3_node2 = nodes[6].name

#     assert final_node == "function_multiple_diverse_type_params"
#     assert p1_node == "p1.node_1"
#     assert p2_node1 == "p2.node_1"
#     assert p2_node2 == "p2.node_2"
#     assert p2_node3 == "p2.node_3"
#     assert p3_node1 == "p3.node_1"
#     assert p3_node2 == "p3.node_3"


# def test_pipe_input_custom_namespace_with_target():
#     n = node.Node.from_fn(function_multiple_diverse_type_params)

#     decorator = pipe_input(
#         step(_test_apply_function, source("bar_upstream"), baz=value(10))
#         .on_input(["p1", "p3"])
#         .named("node_1"),
#         step(_test_apply_function, source("bar_upstream"), baz=value(100)).named("node_2"),
#         step(_test_apply_function, source("bar_upstream"), baz=value(1000))
#         .on_input("p3")
#         .named("node_3"),
#         namespace="abc",
#         on_input="p2",
#     )
#     nodes = decorator.transform_dag([n], {}, function_multiple_diverse_type_params)
#     final_node = nodes[0].name
#     p1_node = nodes[1].name
#     p2_node1 = nodes[2].name
#     p2_node2 = nodes[3].name
#     p2_node3 = nodes[4].name
#     p3_node1 = nodes[5].name
#     p3_node2 = nodes[6].name

#     assert final_node == "function_multiple_diverse_type_params"
#     assert p1_node == "abc_p1.node_1"
#     assert p2_node1 == "abc_p2.node_1"
#     assert p2_node2 == "abc_p2.node_2"
#     assert p2_node3 == "abc_p2.node_3"
#     assert p3_node1 == "abc_p3.node_1"
#     assert p3_node2 == "abc_p3.node_3"


# def test_pipe_input_mapping_args_targets_local():
#     n = node.Node.from_fn(function_multiple_diverse_type_params)

#     decorator = pipe_input(
#         step(_test_apply_function, source("bar_upstream"), baz=value(10))
#         .on_input(["p1", "p3"])
#         .named("node_1"),
#         step(_test_apply_function, source("bar_upstream"), baz=value(100))
#         .on_input("p2")
#         .named("node_2"),
#         step(_test_apply_function, source("bar_upstream"), baz=value(1000))
#         .on_input("p3")
#         .named("node_3"),
#         namespace="abc",
#     )
#     nodes = decorator.transform_dag([n], {}, function_multiple_diverse_type_params)
#     nodes_by_name = {item.name: item for item in nodes}
#     chain_node_1 = nodes_by_name["abc_p1.node_1"]
#     chain_node_2 = nodes_by_name["abc_p2.node_2"]
#     chain_node_3_first = nodes_by_name["abc_p3.node_1"]
#     assert chain_node_1(p1=1, bar_upstream=3) == 14
#     assert chain_node_2(p2=1, bar_upstream=3) == 104
#     assert chain_node_3_first(p3=7, bar_upstream=3) == 20
#
#
# def test_pipe_input_mapping_args_targets_local_adds_to_global():
#     n = node.Node.from_fn(function_multiple_same_type_params)

#     decorator = pipe_input(
#         step(_test_apply_function, source("bar_upstream"), baz=value(10))
#         .on_input(["p1", "p2"])
#         .named("node_1"),
#         step(_test_apply_function, source("bar_upstream"), baz=value(100))
#         .on_input("p2")
#         .named("node_2"),
#         step(_test_apply_function, source("bar_upstream"), baz=value(1000)).named("node_3"),
#         on_input="p3",
#         namespace="abc",
#     )
#     nodes = decorator.transform_dag([n], {}, function_multiple_diverse_type_params)
#     nodes_by_name = {item.name: item for item in nodes}
#     p1_node = nodes_by_name["abc_p1.node_1"]
#     p2_node1 = nodes_by_name["abc_p2.node_1"]
#     p2_node2 = nodes_by_name["abc_p2.node_2"]
#     p3_node1 = nodes_by_name["abc_p3.node_1"]
#     p3_node2 = nodes_by_name["abc_p3.node_2"]
#     p3_node3 = nodes_by_name["abc_p3.node_3"]

#     assert p1_node(p1=1, bar_upstream=3) == 14
#     assert p2_node1(p2=7, bar_upstream=3) == 20
#     assert p2_node2(**{"abc_p2.node_1": 2, "bar_upstream": 3}) == 105
#     assert p3_node1(p3=9, bar_upstream=3) == 22
#     assert p3_node2(**{"abc_p3.node_1": 13, "bar_upstream": 3}) == 116
#     assert p3_node3(**{"abc_p3.node_2": 17, "bar_upstream": 3}) == 1020


# def test_pipe_input_fails_with_missing_targets():
#     n = node.Node.from_fn(function_multiple_same_type_params)

#     decorator = pipe_input(
#         step(_test_apply_function, source("bar_upstream"), baz=value(10))
#         .on_input(["p1", "p2"])
#         .named("node_1"),
#         step(_test_apply_function, source("bar_upstream"), baz=value(100))
#         .on_input("p2")
#         .named("node_2"),
#         step(_test_apply_function, source("bar_upstream"), baz=value(1000)).named("node_3"),
#         namespace="abc",
#     )
#     with pytest.raises(hamilton.function_modifiers.macros.MissingTargetError):
#         nodes = decorator.transform_dag([n], {}, function_multiple_same_type_params)  # noqa


# def test_pipe_input_decorator_with_target_no_collapse_multi_node():
#     n = node.Node.from_fn(function_multiple_same_type_params)

#     decorator = pipe_input(
#         step(_test_apply_function, source("bar_upstream"), baz=value(10))
#         .on_input(["p1", "p3"])
#         .named("node_1"),
#         step(_test_apply_function, source("bar_upstream"), baz=value(100))
#         .on_input("p2")
#         .named("node_2"),
#         step(_test_apply_function, source("bar_upstream"), baz=value(1000))
#         .on_input("p3")
#         .named("node_3"),
#         namespace="abc",
#     )
#     nodes = decorator.transform_dag([n], {}, function_multiple_diverse_type_params)
#     nodes_by_name = {item.name: item for item in nodes}
#     final_node = nodes_by_name["function_multiple_same_type_params"]
#     chain_node_1 = nodes_by_name["abc_p1.node_1"]
#     chain_node_2 = nodes_by_name["abc_p2.node_2"]
#     chain_node_3_first = nodes_by_name["abc_p3.node_1"]
#     chain_node_3_second = nodes_by_name["abc_p3.node_3"]
#     assert len(nodes_by_name) == 5
#     assert chain_node_1(p1=1, bar_upstream=3) == 14
#     assert chain_node_2(p2=1, bar_upstream=3) == 104
#     assert chain_node_3_first(p3=7, bar_upstream=3) == 20
#     assert chain_node_3_second(**{"abc_p3.node_1": 13, "bar_upstream": 3}) == 1016
#     assert final_node(**{"abc_p1.node_1": 3, "abc_p2.node_2": 4, "abc_p3.node_3": 5}) == 12


def test_pipe_decorator_positional_variable_args():
    n = node.Node.from_fn(general_downstream_function)

    decorator = pipe_input(
        step(_test_apply_function, source("bar_upstream"), baz=value(1000)).named("node_1"),
        namespace=None,
    )
    nodes = decorator.transform_dag([n], {}, general_downstream_function)
    nodes_by_name = {item.name: item for item in nodes}
    chain_node = nodes_by_name["node_1"]
    assert chain_node(result=1, bar_upstream=10) == 1011  # This chains it through
    assert sorted(chain_node.input_types) == ["bar_upstream", "result"]
    final_node = nodes_by_name["general_downstream_function"]
    assert final_node(node_1=1) == 1


def test_pipe_decorator_no_collapse_multi_node():
    n = node.Node.from_fn(general_downstream_function)

    decorator = pipe_input(
        step(_test_apply_function, bar=source("bar_upstream"), baz=100).named("node_1"),
        step(_test_apply_function, bar=value(10), baz=value(100)).named("node_2"),
        namespace=None,
    )
    nodes = decorator.transform_dag([n], {}, general_downstream_function)
    nodes_by_name = {item.name: item for item in nodes}
    final_node = nodes_by_name["general_downstream_function"]
    assert len(nodes_by_name) == 3
    assert nodes_by_name["node_1"](result=1, bar_upstream=10) == 111
    assert nodes_by_name["node_2"](node_1=1) == 111
    assert final_node(node_2=100) == 100


def test_resolve_namespace_inherit():
    applicable = Applicable(
        _test_apply_function, args=(), kwargs=dict(bar=source("bar_upstream"), baz=100)
    ).named("node_1")
    assert applicable.resolve_namespace("inherited") == ("inherited",)


def test_resolve_namespace_discard():
    applicable = Applicable(
        _test_apply_function, args=(), kwargs=dict(bar=source("bar_upstream"), baz=100)
    ).named("node_1", namespace=None)
    assert applicable.resolve_namespace("unused") == ()


def test_resolve_namespace_replaced():
    applicable = Applicable(
        _test_apply_function, args=(), kwargs=dict(bar=source("bar_upstream"), baz=100)
    ).named("node_1", namespace="replaced")
    assert applicable.resolve_namespace("unused") == ("replaced",)


def test_validate_pipe_fails_with_conflicting_namespace():
    decorator = pipe_input(
        step(_test_apply_function, bar=source("bar_upstream"), baz=100).named(
            "node_1", namespace="custom"
        ),
        namespace=None,  # Not allowed to have custom namespacess if the namespace is None
    )
    with pytest.raises(hamilton.function_modifiers.base.InvalidDecoratorException):
        decorator.validate(general_downstream_function)


def test_inherits_null_namespace():
    n = node.Node.from_fn(general_downstream_function)
    decorator = pipe_input(
        step(_test_apply_function, bar=source("bar_upstream"), baz=100).named(
            "node_1", namespace=...
        ),
        namespace=None,  # Not allowed to have custom namespacess if the namespace is None
    )
    decorator.validate(general_downstream_function)
    nodes = decorator.transform_dag([n], {}, general_downstream_function)
    assert "node_1" in {item.name for item in nodes}
    assert "general_downstream_function" in {item.name for item in nodes}


def test_pipe_end_to_end_1():
    dr = (
        driver.Builder()
        .with_modules(tests.resources.pipe_input)
        .with_adapter(base.DefaultAdapter())
        .with_config({"calc_c": True})
        .build()
    )

    inputs = {
        "input_1": 10,
        "input_2": 20,
        "input_3": 30,
    }
    result = dr.execute(
        [
            "chain_1_using_pipe",
            "chain_2_using_pipe",
            "chain_1_not_using_pipe",
            "chain_2_not_using_pipe",
        ],
        inputs=inputs,
    )
    assert result["chain_1_using_pipe"] == result["chain_1_not_using_pipe"]
    assert result["chain_2_using_pipe"] == result["chain_2_not_using_pipe"]


def test_pipe_end_to_end_target_global():
    dr = (
        driver.Builder()
        .with_modules(tests.resources.pipe_input)
        .with_adapter(base.DefaultAdapter())
        .with_config({"calc_c": True})
        .build()
    )

    inputs = {
        "input_1": 10,
        "input_2": 20,
        "input_3": 30,
    }
    result = dr.execute(
        [
            "chain_1_using_pipe_input_target_global",
            "chain_1_not_using_pipe_input_target_global",
        ],
        inputs=inputs,
    )
    assert (
        result["chain_1_not_using_pipe_input_target_global"]
        == result["chain_1_using_pipe_input_target_global"]
    )


# TODO: For multiple parameters end-to-end
# def test_pipe_end_to_end_target_local():
#     dr = (
#         driver.Builder()
#         .with_modules(tests.resources.pipe_input)
#         .with_adapter(base.DefaultAdapter())
#         .with_config({"calc_c": True})
#         .build()
#     )

#     inputs = {
#         "input_1": 10,
#         "input_2": 20,
#         "input_3": 30,
#     }
#     result = dr.execute(
#         [
#             "chain_1_using_pipe_input_target_local",
#             "chain_1_not_using_pipe_input_target_local",
#         ],
#         inputs=inputs,
#     )
#     assert (
#         result["chain_1_not_using_pipe_input_target_local"]
#         == result["chain_1_using_pipe_input_target_local"]
#     )

# def test_pipe_end_to_end_target_mixed():
#     dr = (
#         driver.Builder()
#         .with_modules(tests.resources.pipe_input)
#         .with_adapter(base.DefaultAdapter())
#         .with_config({"calc_c": True})
#         .build()
#     )

#     inputs = {
#         "input_1": 10,
#         "input_2": 20,
#         "input_3": 30,
#     }
#     result = dr.execute(
#         [
#             "chain_1_using_pipe_input_target_mixed",
#             "chain_1_not_using_pipe_input_target_mixed",
#         ],
#         inputs=inputs,
#     )
#     assert (
#         result["chain_1_not_using_pipe_input_target_mixed"]
#         == result["chain_1_using_pipe_input_target_mixed"]
#     )


def result_from_downstream_function() -> int:
    return 2


def test_pipe_output_single_target_level_error():
    with pytest.raises(hamilton.function_modifiers.macros.SingleTargetError):
        pipe_output(
            step(_test_apply_function, source("bar_upstream"), baz=value(100)).on_output(
                "some_node"
            ),
            on_output="some_other_node",
        )


def test_pipe_output_shortcircuit():
    n = node.Node.from_fn(result_from_downstream_function)
    decorator = pipe_output()
    nodes = decorator.transform_dag([n], {}, result_from_downstream_function)
    assert len(nodes) == 1
    assert n == nodes[0]


def test_pipe_output_decorator_positional_single_node():
    n = node.Node.from_fn(result_from_downstream_function)

    decorator = pipe_output(
        step(_test_apply_function, source("bar_upstream"), baz=value(100)).named("node_1"),
        namespace=None,
    )
    nodes = decorator.transform_dag([n], {}, result_from_downstream_function)
    nodes_by_name = {item.name: item for item in nodes}
    chain_node = nodes_by_name["node_1"]
    assert chain_node(**{"result_from_downstream_function.raw": 2, "bar_upstream": 10}) == 112
    assert sorted(chain_node.input_types) == [
        "bar_upstream",
        "result_from_downstream_function.raw",
    ]
    final_node = nodes_by_name["result_from_downstream_function"]
    assert final_node(foo=112) == 112  # original arg name
    assert final_node(node_1=112) == 112  # renamed to match the last node


def test_pipe_output_decorator_no_collapse_multi_node():
    n = node.Node.from_fn(result_from_downstream_function)

    decorator = pipe_output(
        step(_test_apply_function, bar=source("bar_upstream"), baz=100).named("node_1"),
        step(_test_apply_function, bar=value(10), baz=value(100)).named("node_2"),
        namespace=None,
    )
    nodes = decorator.transform_dag([n], {}, result_from_downstream_function)
    nodes_by_name = {item.name: item for item in nodes}
    final_node = nodes_by_name["result_from_downstream_function"]
    assert len(nodes_by_name) == 4  # We add fn_raw and identity
    assert (
        nodes_by_name["node_1"](**{"result_from_downstream_function.raw": 1, "bar_upstream": 10})
        == 111
    )
    assert nodes_by_name["node_2"](node_1=4) == 114
    assert final_node(node_2=13) == 13


def test_validate_pipe_output_fails_with_conflicting_namespace():
    decorator = pipe_output(
        step(_test_apply_function, bar=source("bar_upstream"), baz=100).named(
            "node_1", namespace="custom"
        ),
        namespace=None,  # Not allowed to have custom namespacess if the namespace is None
    )
    with pytest.raises(hamilton.function_modifiers.base.InvalidDecoratorException):
        decorator.validate(result_from_downstream_function)


def test_pipe_output_inherits_null_namespace():
    n = node.Node.from_fn(result_from_downstream_function)
    decorator = pipe_output(
        step(_test_apply_function, bar=source("bar_upstream"), baz=100).named(
            "node_1", namespace=...
        ),
        namespace=None,  # Not allowed to have custom namespacess if the namespace is None
    )
    decorator.validate(result_from_downstream_function)
    nodes = decorator.transform_dag([n], {}, result_from_downstream_function)
    assert "node_1" in {item.name for item in nodes}
    assert "result_from_downstream_function.raw" in {item.name for item in nodes}
    assert "result_from_downstream_function" in {item.name for item in nodes}


def test_pipe_output_global_on_output_all():
    n1 = node.Node.from_fn(result_from_downstream_function, name="node_1")
    n2 = node.Node.from_fn(result_from_downstream_function, name="node_2")

    decorator = pipe_output(
        step(_test_apply_function, source("bar_upstream"), baz=value(100)),
    )
    nodes = decorator.select_nodes(decorator.target, [n1, n2])
    assert len(nodes) == 2
    assert [node_.name for node_ in nodes] == ["node_1", "node_2"]


def test_pipe_output_global_on_output_string():
    n1 = node.Node.from_fn(result_from_downstream_function, name="node_1")
    n2 = node.Node.from_fn(result_from_downstream_function, name="node_2")

    decorator = pipe_output(
        step(_test_apply_function, source("bar_upstream"), baz=value(100)), on_output="node_2"
    )
    nodes = decorator.select_nodes(decorator.target, [n1, n2])
    assert len(nodes) == 1
    assert nodes[0].name == "node_2"


def test_pipe_output_global_on_output_list_strings():
    n1 = node.Node.from_fn(result_from_downstream_function, name="node_1")
    n2 = node.Node.from_fn(result_from_downstream_function, name="node_2")
    n3 = node.Node.from_fn(result_from_downstream_function, name="node_3")

    decorator = pipe_output(
        step(_test_apply_function, source("bar_upstream"), baz=value(100)),
        on_output=["node_1", "node_2"],
    )
    nodes = decorator.select_nodes(decorator.target, [n1, n2, n3])
    assert len(nodes) == 2
    assert [node_.name for node_ in nodes] == ["node_1", "node_2"]


def test_pipe_output_elipsis_error():
    with pytest.raises(ValueError):
        pipe_output(
            step(_test_apply_function, source("bar_upstream"), baz=value(100)), on_output=...
        )


def test_pipe_output_local_on_output_string():
    n1 = node.Node.from_fn(result_from_downstream_function, name="node_1")
    n2 = node.Node.from_fn(result_from_downstream_function, name="node_2")

    decorator = pipe_output(
        step(_test_apply_function, source("bar_upstream"), baz=value(100))
        .named("correct_transform")
        .on_output("node_2"),
        step(_test_apply_function, source("bar_upstream"), baz=value(100))
        .named("wrong_transform")
        .on_output("node_3"),
    )
    steps = decorator._filter_individual_target(n1)
    assert len(steps) == 0
    steps = decorator._filter_individual_target(n2)
    assert len(steps) == 1
    assert steps[0].name == "correct_transform"


def test_pipe_output_local_on_output_list_string():
    n1 = node.Node.from_fn(result_from_downstream_function, name="node_1")
    n2 = node.Node.from_fn(result_from_downstream_function, name="node_2")
    n3 = node.Node.from_fn(result_from_downstream_function, name="node_3")

    decorator = pipe_output(
        step(_test_apply_function, source("bar_upstream"), baz=value(100))
        .named("correct_transform_list")
        .on_output(["node_2", "node_3"]),
        step(_test_apply_function, source("bar_upstream"), baz=value(100))
        .named("correct_transform_string")
        .on_output("node_2"),
        step(_test_apply_function, source("bar_upstream"), baz=value(100))
        .named("wrong_transform")
        .on_output("node_5"),
    )
    steps = decorator._filter_individual_target(n1)
    assert len(steps) == 0
    steps = decorator._filter_individual_target(n2)
    assert len(steps) == 2
    assert steps[0].name == "correct_transform_list"
    assert steps[1].name == "correct_transform_string"
    steps = decorator._filter_individual_target(n3)
    assert len(steps) == 1
    assert steps[0].name == "correct_transform_list"


def test_pipe_output_end_to_end_simple():
    dr = driver.Builder().with_config({"calc_c": True}).build()

    dr = (
        driver.Builder()
        .with_modules(tests.resources.pipe_output)
        .with_adapter(base.DefaultAdapter())
        .build()
    )

    inputs = {}
    result = dr.execute(
        [
            "downstream_f",
            "chain_not_using_pipe_output",
        ],
        inputs=inputs,
    )
    assert result["downstream_f"] == result["chain_not_using_pipe_output"]


def test_pipe_output_end_to_end():
    dr = (
        driver.Builder()
        .with_modules(tests.resources.pipe_output)
        .with_adapter(base.DefaultAdapter())
        .with_config({"calc_c": True})
        .build()
    )

    inputs = {
        "input_1": 10,
        "input_2": 20,
        "input_3": 30,
    }
    result = dr.execute(
        [
            "chain_1_using_pipe_output",
            "chain_2_using_pipe_output",
            "chain_1_not_using_pipe_output",
            "chain_2_not_using_pipe_output",
        ],
        inputs=inputs,
    )
    assert result["chain_1_using_pipe_output"] == result["chain_1_not_using_pipe_output"]
    assert result["chain_2_using_pipe_output"] == result["chain_2_not_using_pipe_output"]


# Mutate will mark the modules (and leave a mark).
# Thus calling it a second time (for instance through pmultiple tests) might mess it up slightly...
# Using fixtures just to be sure.


@pytest.fixture(scope="function")
def _downstream_result_to_mutate():
    def downstream_result_to_mutate() -> int:
        return 2

    yield downstream_result_to_mutate


@pytest.fixture(scope="function")
def import_mutate_module():
    import importlib

    mod = importlib.import_module("tests.resources.mutate")
    yield mod


# This doesn't change so no need to have it as fixture
def mutator_function(input_1: int, input_2: int) -> int:
    return input_1 + input_2


def test_mutate_convert_callable_to_applicable(_downstream_result_to_mutate):
    decorator = mutate(_downstream_result_to_mutate)

    assert len(decorator.remote_applicables) == 1
    remote_applicable = decorator.remote_applicables[0]
    assert isinstance(remote_applicable, Applicable)
    assert remote_applicable.fn is None
    assert remote_applicable.target_fn == _downstream_result_to_mutate


def test_mutate_restricted_to_same_module():
    decorator = mutate(tests.resources.mutate.f_of_interest)

    with pytest.raises(hamilton.function_modifiers.macros.NotSameModuleError):
        decorator.validate_same_module(mutator_function)


def test_mutate_global_kwargs(_downstream_result_to_mutate):
    decorator = mutate(apply_to(_downstream_result_to_mutate), input_2=17)
    remote_applicable = decorator.remote_applicables[0]

    pipe_step = decorator._create_step(
        mutating_fn=mutator_function, remote_applicable_builder=remote_applicable
    )
    assert pipe_step.kwargs["input_2"] == 17


def test_mutate_local_kwargs_override_global_ones(_downstream_result_to_mutate):
    decorator = mutate(apply_to(_downstream_result_to_mutate, input_2=13), input_2=17)
    remote_applicable = decorator.remote_applicables[0]

    pipe_step = decorator._create_step(
        mutating_fn=mutator_function, remote_applicable_builder=remote_applicable
    )
    assert pipe_step.kwargs["input_2"] == 13


def test_mutate_end_to_end_simple(import_mutate_module):
    dr = driver.Builder().with_config({"calc_c": True}).build()

    dr = (
        driver.Builder()
        .with_modules(import_mutate_module)
        .with_adapter(base.DefaultAdapter())
        .build()
    )

    inputs = {}
    result = dr.execute(
        [
            "downstream_f",
            "chain_not_using_mutate",
        ],
        inputs=inputs,
    )
    assert result["downstream_f"] == result["chain_not_using_mutate"]


def test_mutate_end_to_end_1(import_mutate_module):
    dr = (
        driver.Builder()
        .with_modules(import_mutate_module)
        .with_adapter(base.DefaultAdapter())
        .with_config({"calc_c": True})
        .build()
    )

    inputs = {
        "input_1": 10,
        "input_2": 20,
        "input_3": 30,
    }
    result = dr.execute(
        [
            "chain_1_using_mutate",
            "chain_2_using_mutate",
            "chain_1_not_using_mutate",
            "chain_2_not_using_mutate",
        ],
        inputs=inputs,
    )
    assert result["chain_1_using_mutate"] == result["chain_1_not_using_mutate"]
    assert result["chain_2_using_mutate"] == result["chain_2_not_using_mutate"]
